from abc import ABC, abstractmethod
import math
import warnings

import pandas as pd

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch import Tensor

from netam.hit_class import apply_multihit_correction
from netam.common import (
    MAX_AMBIG_AA_IDX,
    aa_idx_tensor_of_str_ambig,
    PositionalEncoding,
    generate_kmers,
    aa_mask_tensor_of,
    encode_sequences,
    chunk_method,
)

warnings.filterwarnings(
    "ignore", category=UserWarning, module="torch.nn.modules.transformer"
)


class ModelBase(nn.Module):
    def reinitialize_weights(self):
        for layer in self.children():
            if isinstance(layer, nn.Embedding):
                nn.init.normal_(layer.weight)
            elif isinstance(layer, nn.Linear):
                nn.init.kaiming_normal_(layer.weight, nonlinearity="linear")
                if layer.bias is not None:
                    nn.init.constant_(layer.bias, 0)
            elif isinstance(layer, nn.Conv1d):
                nn.init.kaiming_normal_(layer.weight, nonlinearity="relu")
                if layer.bias is not None:
                    nn.init.constant_(layer.bias, 0)
            elif isinstance(layer, nn.TransformerEncoder):
                for sublayer in layer.modules():
                    if isinstance(sublayer, nn.Linear):
                        nn.init.kaiming_normal_(sublayer.weight, nonlinearity="relu")
                        if sublayer.bias is not None:
                            nn.init.constant_(sublayer.bias, 0)
            elif isinstance(layer, nn.Dropout):
                pass
            elif hasattr(layer, "reinitialize_weights"):
                layer.reinitialize_weights()
            else:
                raise ValueError(f"Unrecognized layer type: {type(layer)}")

    @property
    def device(self):
        return next(self.parameters()).device

    def freeze(self):
        """Freeze all parameters in the model, disabling gradient computations."""
        for param in self.parameters():
            param.requires_grad = False

    def unfreeze(self):
        """Unfreeze all parameters in the model, enabling gradient computations."""
        for param in self.parameters():
            param.requires_grad = True

    @chunk_method(progress_bar_name="Evaluating model")
    def evaluate_sequences(self, sequences, encoder=None, chunk_size=2048):
        if encoder is None:
            raise ValueError("An encoder must be provided.")
        encoded_parents, masks, wt_base_modifiers = encode_sequences(sequences, encoder)
        encoded_parents = encoded_parents.to(self.device)
        masks = masks.to(self.device)
        wt_base_modifiers = wt_base_modifiers.to(self.device)
        with torch.no_grad():
            outputs = self(encoded_parents, masks, wt_base_modifiers)
            return tuple(t.detach().cpu() for t in outputs)


class KmerModel(ModelBase):
    def __init__(self, kmer_length):
        super().__init__()
        self.kmer_length = kmer_length
        self.all_kmers = generate_kmers(kmer_length)
        self.kmer_count = len(self.all_kmers)

    @property
    def hyperparameters(self):
        return {
            "kmer_length": self.kmer_length,
        }


class FivemerModel(KmerModel):
    def __init__(self):
        super().__init__(kmer_length=5)
        self.kmer_embedding = nn.Embedding(self.kmer_count, 1)

    def forward(self, encoded_parents, masks, wt_base_modifier):
        log_kmer_rates = self.kmer_embedding(encoded_parents).squeeze(-1)
        rates = torch.exp(log_kmer_rates * masks)
        return rates

    def adjust_rate_bias_by(self, log_adjustment_factor):
        with torch.no_grad():
            self.kmer_embedding.weight.data += log_adjustment_factor

    @property
    def kmer_rates(self):
        return torch.exp(self.kmer_embedding.weight).squeeze()


class RSFivemerModel(KmerModel):
    def __init__(self, kmer_length=5):
        assert kmer_length == 5
        super().__init__(kmer_length=5)
        self.r_kmer_embedding = nn.Embedding(self.kmer_count, 1)
        self.s_kmer_embedding = nn.Embedding(self.kmer_count, 4)

    def forward(self, encoded_parents, masks, wt_base_modifier):
        log_kmer_rates = self.r_kmer_embedding(encoded_parents).squeeze(-1)
        rates = torch.exp(log_kmer_rates * masks)
        csp_logits = self.s_kmer_embedding(encoded_parents)
        # When we have an N, set all the CSP logits to 0, resulting in a uniform
        # prediction. There is nothing to predict here.
        csp_logits *= masks.unsqueeze(-1)
        # As described elsewhere, this makes the WT base have a probability of 0
        # after softmax.
        csp_logits += wt_base_modifier
        return rates, csp_logits

    def adjust_rate_bias_by(self, log_adjustment_factor):
        with torch.no_grad():
            self.r_kmer_embedding.weight.data += log_adjustment_factor

    @property
    def kmer_rates(self):
        return torch.exp(self.r_kmer_embedding.weight).squeeze()


class SHMoofModel(KmerModel):
    def __init__(self, kmer_length, site_count):
        super().__init__(kmer_length)
        self.site_count = site_count
        self.kmer_embedding = nn.Embedding(self.kmer_count, 1)
        self.log_site_rates = nn.Embedding(self.site_count, 1)

    def forward(self, encoded_parents, masks, wt_base_modifier):
        log_kmer_rates = self.kmer_embedding(encoded_parents).squeeze(-1)
        # When we transpose we get a tensor of shape [site_count, 1], which will broadcast
        # to the shape of log_kmer_rates, repeating over the batch dimension.
        log_site_rates = self.log_site_rates.weight.T
        # Rates are the product of kmer and site rates.
        rates = torch.exp((log_kmer_rates + log_site_rates) * masks)
        return rates

    def adjust_rate_bias_by(self, log_adjustment_factor):
        with torch.no_grad():
            self.kmer_embedding.weight.data += log_adjustment_factor / 2.0
            self.log_site_rates.weight.data += log_adjustment_factor / 2.0

    @property
    def hyperparameters(self):
        return {
            "kmer_length": self.kmer_length,
            "site_count": self.site_count,
        }

    @property
    def kmer_rates(self):
        return torch.exp(self.kmer_embedding.weight).squeeze()

    @property
    def site_rates(self):
        return torch.exp(self.log_site_rates.weight).squeeze()

    def write_shmoof_output(self, out_dir):
        # Extract k-mer (motif) mutabilities
        kmer_rates = self.kmer_rates.detach().numpy().flatten()
        motif_mutabilities = pd.DataFrame(
            {
                "Motif": self.all_kmers,
                "Mutability": kmer_rates,
            }
        )
        motif_mutabilities.to_csv(
            f"{out_dir}/motif_mutabilities.tsv", sep="\t", index=False
        )

        # Extract site mutabilities
        site_mutabilities = self.site_rates.detach().numpy().flatten()
        site_mutabilities_df = pd.DataFrame(
            {
                "Position": range(1, len(site_mutabilities) + 1),
                "Mutability": site_mutabilities,
            }
        )
        site_mutabilities_df.to_csv(
            f"{out_dir}/site_mutabilities.tsv", sep="\t", index=False
        )


class RSSHMoofModel(KmerModel):
    def __init__(self, kmer_length, site_count):
        super().__init__(kmer_length)
        self.site_count = site_count
        self.kmer_embedding = nn.Embedding(self.kmer_count, 4)
        self.log_site_rates = nn.Embedding(self.site_count, 1)

    def forward(self, encoded_parents, masks, wt_base_modifier):
        log_kmer_rates_per_base = self.kmer_embedding(encoded_parents)
        # Set WT base to have rate of 0 in log space.
        log_kmer_rates_per_base += wt_base_modifier
        # Here we are summing over the per-base dimensions to get the per-kmer
        # rates. We want to do so in linear, not log space.
        log_kmer_rates = torch.logsumexp(log_kmer_rates_per_base, dim=-1)
        assert log_kmer_rates.shape == (
            encoded_parents.size(0),
            encoded_parents.size(1),
        )

        # When we transpose we get a tensor of shape [site_count, 1], which will broadcast
        # to the shape of log_kmer_rates, repeating over the batch dimension.
        log_site_rates = self.log_site_rates.weight.T
        # Rates are the product of kmer and site rates.
        rates = torch.exp((log_kmer_rates + log_site_rates) * masks)

        csp_logits = log_kmer_rates_per_base * masks.unsqueeze(-1)
        csp_logits += wt_base_modifier

        return rates, csp_logits

    def adjust_rate_bias_by(self, log_adjustment_factor):
        with torch.no_grad():
            self.kmer_embedding.weight.data += log_adjustment_factor / 2.0
            self.log_site_rates.weight.data += log_adjustment_factor / 2.0

    @property
    def hyperparameters(self):
        return {
            "kmer_length": self.kmer_length,
            "site_count": self.site_count,
        }


class CNNModel(KmerModel):
    """This is a CNN model that uses k-mers as input and trains an embedding layer."""

    def __init__(
        self, kmer_length, embedding_dim, filter_count, kernel_size, dropout_prob=0.1
    ):
        super().__init__(kmer_length)
        self.kmer_embedding = nn.Embedding(self.kmer_count, embedding_dim)
        self.conv = nn.Conv1d(
            in_channels=embedding_dim,
            out_channels=filter_count,
            kernel_size=kernel_size,
            padding="same",
        )
        self.dropout = nn.Dropout(dropout_prob)
        self.linear = nn.Linear(in_features=filter_count, out_features=1)

    def forward(self, encoded_parents, masks, wt_base_modifier):
        kmer_embeds = self.kmer_embedding(encoded_parents)
        kmer_embeds = kmer_embeds.permute(0, 2, 1)  # Transpose for Conv1D
        conv_out = F.relu(self.conv(kmer_embeds))
        conv_out = self.dropout(conv_out)
        conv_out = conv_out.permute(0, 2, 1)  # Transpose back for Linear layer
        log_rates = self.linear(conv_out).squeeze(-1)
        rates = torch.exp(log_rates * masks)
        return rates

    def adjust_rate_bias_by(self, log_adjustment_factor):
        with torch.no_grad():
            self.linear.bias.data += log_adjustment_factor

    @property
    def hyperparameters(self):
        return {
            "kmer_length": self.kmer_length,
            "embedding_dim": self.kmer_embedding.embedding_dim,
            "filter_count": self.conv.out_channels,
            "kernel_size": self.conv.kernel_size[0],
            "dropout_prob": self.dropout.p,
        }


class CNNPEModel(CNNModel):
    def __init__(
        self, kmer_length, embedding_dim, filter_count, kernel_size, dropout_prob
    ):
        super().__init__(
            kmer_length, embedding_dim, filter_count, kernel_size, dropout_prob
        )
        self.pos_encoder = PositionalEncoding(embedding_dim, dropout=dropout_prob)

    def forward(self, encoded_parents, masks, wt_base_modifier):
        kmer_embeds = self.kmer_embedding(encoded_parents)
        kmer_embeds = self.pos_encoder(kmer_embeds)
        kmer_embeds = kmer_embeds.permute(0, 2, 1)  # Transpose for Conv1D
        conv_out = F.relu(self.conv(kmer_embeds))
        conv_out = self.dropout(conv_out)
        conv_out = conv_out.permute(0, 2, 1)  # Transpose back for Linear layer
        log_rates = self.linear(conv_out).squeeze(-1)
        rates = torch.exp(log_rates * masks)
        return rates


class CNN1merModel(CNNModel):
    """This is a CNN model that uses individual bases as input and does not train an
    embedding layer."""

    def __init__(self, filter_count, kernel_size, dropout_prob=0.1):
        # Fixed embedding_dim because there are only 4 bases.
        embedding_dim = 5
        kmer_length = 1
        super().__init__(
            kmer_length, embedding_dim, filter_count, kernel_size, dropout_prob
        )
        # Here's how we adapt the model to use individual bases as input rather
        # than trainable kmer embeddings.
        identity_matrix = torch.eye(embedding_dim)
        self.kmer_embedding.weight = nn.Parameter(identity_matrix, requires_grad=False)


class RSCNNModel(CNNModel, ABC):
    """The base class for all RSCNN models. These are CNN models that predict both rates
    and CSP logits.

    They differ in how much they share the weights between the r_ components that
    predict rates, and the s_ components that predict CSP logits.

    See
    https://github.com/matsengrp/netam/pull/9#issuecomment-1939097576
    for diagrams about the various models.
    """

    @abstractmethod
    def forward(self, encoded_parents, masks, wt_base_modifier):
        pass

    def adjust_rate_bias_by(self, log_adjustment_factor):
        with torch.no_grad():
            self.r_linear.bias.data += log_adjustment_factor


class JoinedRSCNNModel(RSCNNModel):
    """This model shares everything except the final linear layers."""

    def __init__(
        self, kmer_length, embedding_dim, filter_count, kernel_size, dropout_prob=0.1
    ):
        super().__init__(
            kmer_length, embedding_dim, filter_count, kernel_size, dropout_prob
        )
        self.kmer_embedding = nn.Embedding(self.kmer_count, embedding_dim)
        self.conv = nn.Conv1d(
            in_channels=embedding_dim,
            out_channels=filter_count,
            kernel_size=kernel_size,
            padding="same",
        )
        self.dropout = nn.Dropout(dropout_prob)
        self.r_linear = nn.Linear(in_features=filter_count, out_features=1)
        self.s_linear = nn.Linear(in_features=filter_count, out_features=4)

    def forward(self, encoded_parents, masks, wt_base_modifier):
        kmer_embeds = self.kmer_embedding(encoded_parents)
        kmer_embeds = kmer_embeds.permute(0, 2, 1)  # Transpose for Conv1D
        conv_out = F.relu(self.conv(kmer_embeds))
        conv_out = self.dropout(conv_out)
        conv_out = conv_out.permute(0, 2, 1)

        log_rates = self.r_linear(conv_out).squeeze(-1)
        rates = torch.exp(log_rates * masks)

        csp_logits = self.s_linear(conv_out)
        csp_logits *= masks.unsqueeze(-1)
        csp_logits += wt_base_modifier

        return rates, csp_logits


class HybridRSCNNModel(RSCNNModel):
    """This model shares the kmer_embedding only."""

    def __init__(
        self, kmer_length, embedding_dim, filter_count, kernel_size, dropout_prob=0.1
    ):
        super().__init__(
            kmer_length, embedding_dim, filter_count, kernel_size, dropout_prob
        )

        self.kmer_embedding = nn.Embedding(self.kmer_count, embedding_dim)
        # Duplicate the layers for the r_ component
        self.r_conv = nn.Conv1d(
            in_channels=embedding_dim,
            out_channels=filter_count,
            kernel_size=kernel_size,
            padding="same",
        )
        self.r_dropout = nn.Dropout(dropout_prob)
        self.r_linear = nn.Linear(in_features=filter_count, out_features=1)

        self.s_conv = nn.Conv1d(
            in_channels=embedding_dim,
            out_channels=filter_count,
            kernel_size=kernel_size,
            padding="same",
        )
        self.s_dropout = nn.Dropout(dropout_prob)
        self.s_linear = nn.Linear(in_features=filter_count, out_features=4)

    def forward(self, encoded_parents, masks, wt_base_modifier):
        kmer_embeds = self.kmer_embedding(encoded_parents)
        kmer_embeds = kmer_embeds.permute(0, 2, 1)
        r_conv_out = F.relu(self.r_conv(kmer_embeds))
        r_conv_out = self.r_dropout(r_conv_out)
        r_conv_out = r_conv_out.permute(0, 2, 1)
        s_conv_out = F.relu(self.s_conv(kmer_embeds))
        s_conv_out = self.s_dropout(s_conv_out)
        s_conv_out = s_conv_out.permute(0, 2, 1)

        log_rates = self.r_linear(r_conv_out).squeeze(-1)
        rates = torch.exp(log_rates * masks)

        csp_logits = self.s_linear(s_conv_out)
        csp_logits *= masks.unsqueeze(-1)
        csp_logits += wt_base_modifier

        return rates, csp_logits


class IndepRSCNNModel(RSCNNModel):
    """This model does not share any weights between the r_ and s_ components."""

    def __init__(
        self, kmer_length, embedding_dim, filter_count, kernel_size, dropout_prob=0.1
    ):
        super().__init__(
            kmer_length, embedding_dim, filter_count, kernel_size, dropout_prob
        )

        # Duplicate the layers for the r_ component
        self.r_kmer_embedding = nn.Embedding(self.kmer_count, embedding_dim)
        self.r_conv = nn.Conv1d(
            in_channels=embedding_dim,
            out_channels=filter_count,
            kernel_size=kernel_size,
            padding="same",
        )
        self.r_dropout = nn.Dropout(dropout_prob)
        self.r_linear = nn.Linear(in_features=filter_count, out_features=1)

        # Duplicate the layers for the s_ component
        self.s_kmer_embedding = nn.Embedding(self.kmer_count, embedding_dim)
        self.s_conv = nn.Conv1d(
            in_channels=embedding_dim,
            out_channels=filter_count,
            kernel_size=kernel_size,
            padding="same",
        )
        self.s_dropout = nn.Dropout(dropout_prob)
        self.s_linear = nn.Linear(in_features=filter_count, out_features=4)

    def forward(self, encoded_parents, masks, wt_base_modifier):
        # Process for r_ component
        r_kmer_embeds = self.r_kmer_embedding(encoded_parents)
        r_kmer_embeds = r_kmer_embeds.permute(0, 2, 1)
        r_conv_out = F.relu(self.r_conv(r_kmer_embeds))
        r_conv_out = self.r_dropout(r_conv_out)
        r_conv_out = r_conv_out.permute(0, 2, 1)

        log_rates = self.r_linear(r_conv_out).squeeze(-1)
        rates = torch.exp(log_rates * masks)

        # Process for s_ component
        s_kmer_embeds = self.s_kmer_embedding(encoded_parents)
        s_kmer_embeds = s_kmer_embeds.permute(0, 2, 1)
        s_conv_out = F.relu(self.s_conv(s_kmer_embeds))
        s_conv_out = self.s_dropout(s_conv_out)
        s_conv_out = s_conv_out.permute(0, 2, 1)

        csp_logits = self.s_linear(s_conv_out)
        csp_logits *= masks.unsqueeze(-1)
        csp_logits += wt_base_modifier

        return rates, csp_logits


# Issue #8
class WrapperHyperparameters:
    def __init__(self, base_model_hyperparameters, site_count):
        self.base_model_hyperparameters = base_model_hyperparameters
        self.site_count = site_count

    def __getitem__(self, key):
        if key in self.base_model_hyperparameters:
            return self.base_model_hyperparameters[key]
        elif hasattr(self, key):
            return getattr(self, key)
        else:
            raise KeyError(f"Key '{key}' not found in hyperparameters.")

    def __str__(self):
        hyperparameters_dict = {key: getattr(self, key) for key in self.__dict__}
        return str(hyperparameters_dict)


class PersiteWrapper(ModelBase):
    """This wraps another model, but adds a per-site rate component."""

    def __init__(self, base_model, site_count):
        super().__init__()
        self.base_model = base_model
        self.site_count = site_count
        self.log_site_rates = nn.Embedding(self.site_count, 1)
        self._hyperparameters = WrapperHyperparameters(
            self.base_model.hyperparameters, self.site_count
        )

    def forward(self, encoded_parents, masks):
        base_model_rates = self.base_model(encoded_parents, masks)
        log_site_rates = self.log_site_rates.weight.T
        rates = base_model_rates * torch.exp(log_site_rates)
        return rates

    @property
    def hyperparameters(self):
        return self._hyperparameters

    @property
    def site_rates(self):
        return torch.exp(self.log_site_rates.weight).squeeze()


class AbstractBinarySelectionModel(ABC, nn.Module):
    """A transformer-based model for binary selection.

    This is a model that takes in a batch of  index-encoded sequences and
    outputs a vector that represents the log level of selection for each amino
    acid site, which after exponentiating is a multiplier on the probability of
    an amino-acid substitution at that site.

    Various submodels are implemented as subclasses of this class:

    * LinAct: No activation function after the transformer.
    * WiggleAct: Activation that slopes to 0 at -inf and grows sub-linearly as x increases.

    See forward() for details.
    """

    def __init__(self):
        super().__init__()

    @property
    def device(self):
        return next(self.parameters()).device

    def evaluate_sequences(self, sequences: list[str], **kwargs) -> Tensor:
        return tuple(self.selection_factors_of_aa_str(seq) for seq in sequences)

    def selection_factors_of_aa_str(self, aa_str: str) -> Tensor:
        """Do the forward method then exponentiation without gradients from an amino
        acid string.

        Args:
            aa_str: A string of amino acids.

        Returns:
            A numpy array of the same length as the input string representing
            the level of selection for each amino acid at each site.
        """

        aa_idxs = aa_idx_tensor_of_str_ambig(aa_str)
        aa_idxs = aa_idxs.to(self.device)
        mask = aa_mask_tensor_of(aa_str)
        mask = mask.to(self.device)

        with torch.no_grad():
            model_out = self(aa_idxs.unsqueeze(0), mask.unsqueeze(0)).squeeze(0)
            final_out = torch.exp(model_out)

        return final_out[: len(aa_str)]


class TransformerBinarySelectionModelLinAct(AbstractBinarySelectionModel):
    def __init__(
        self,
        nhead: int,
        d_model_per_head: int,
        dim_feedforward: int,
        layer_count: int,
        dropout_prob: float = 0.5,
        output_dim: int = 1,
    ):
        super().__init__()
        # Note that d_model has to be divisible by nhead, so we make that
        # automatic here.
        self.d_model_per_head = d_model_per_head
        self.d_model = d_model_per_head * nhead
        self.nhead = nhead
        self.dim_feedforward = dim_feedforward
        self.pos_encoder = PositionalEncoding(self.d_model, dropout_prob)
        self.amino_acid_embedding = nn.Embedding(MAX_AMBIG_AA_IDX + 1, self.d_model)
        self.encoder_layer = nn.TransformerEncoderLayer(
            d_model=self.d_model,
            nhead=nhead,
            dim_feedforward=dim_feedforward,
            batch_first=True,
        )
        self.encoder = nn.TransformerEncoder(self.encoder_layer, layer_count)
        self.linear = nn.Linear(self.d_model, output_dim)
        self.init_weights()

    @property
    def hyperparameters(self):
        return {
            "nhead": self.nhead,
            "d_model_per_head": self.d_model_per_head,
            "dim_feedforward": self.dim_feedforward,
            "layer_count": self.encoder.num_layers,
            "dropout_prob": self.pos_encoder.dropout.p,
            "output_dim": self.linear.out_features,
        }

    def init_weights(self) -> None:
        initrange = 0.1
        self.linear.bias.data.zero_()
        self.linear.weight.data.uniform_(-initrange, initrange)

    def represent(self, amino_acid_indices: Tensor, mask: Tensor) -> Tensor:
        """Represent an index-encoded parent sequence in the model's embedding space.

        Args:
            amino_acid_indices: A tensor of shape (B, L) containing the
                indices of parent AA sequences.
            mask: A tensor of shape (B, L) representing the mask of valid
                amino acid sites.

        Returns:
            The embedded parent sequences, in a tensor of shape (B, L, E),
            where E is the dimensionality of the embedding space.
        """
        # Multiply by sqrt(d_model) to match the transformer paper.
        embedded_amino_acids = self.amino_acid_embedding(
            amino_acid_indices
        ) * math.sqrt(self.d_model)
        # Have to do the permutation because the positional encoding expects the
        # sequence length to be the first dimension.
        embedded_amino_acids = self.pos_encoder(
            embedded_amino_acids.permute(1, 0, 2)
        ).permute(1, 0, 2)

        # To learn about src_key_padding_mask, see https://stackoverflow.com/q/62170439
        return self.encoder(embedded_amino_acids, src_key_padding_mask=~mask)

    def predict(self, representation: Tensor) -> Tensor:
        """Predict selection from the model embedding of a parent sequence.

        Args:
            representation: A tensor of shape (B, L, E) representing the
                embedded parent sequences.
        Returns:
            A tensor of shape (B, L, out_features) representing the log level
            of selection for each amino acid site.
        """
        return self.linear(representation).squeeze(-1)

    def forward(self, amino_acid_indices: Tensor, mask: Tensor) -> Tensor:
        """Build a binary log selection matrix from an index-encoded parent sequence.

        Because we're predicting log of the selection factor, we don't use an
        activation function after the transformer.

        Args:
            amino_acid_indices: A tensor of shape (B, L) containing the indices of parent AA sequences.
            mask: A tensor of shape (B, L) representing the mask of valid amino acid sites.

        Returns:
            A tensor of shape (B, L, out_features) representing the log level
            of selection for each possible amino acid at each site.
        """
        return self.predict(self.represent(amino_acid_indices, mask))


def wiggle(x, beta):
    """A function that when we exp it gives us a function that slopes to 0 at -inf and
    grows sub-linearly as x increases.

    See https://github.com/matsengrp/netam/pull/5#issuecomment-1906665475 for a
    plot.
    """
    return beta * torch.where(x < 1, x - 1, torch.log(x))


class TransformerBinarySelectionModelWiggleAct(TransformerBinarySelectionModelLinAct):
    """Here the beta parameter is fixed at 0.3."""

    def predict(self, representation: Tensor):
        return wiggle(super().predict(representation), 0.3)


class TransformerBinarySelectionModelTrainableWiggleAct(
    TransformerBinarySelectionModelLinAct
):
    """This version of the model has a trainable parameter that controls the beta in the
    wiggle function.

    It didn't work any better so I'm not using it for now.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # Initialize the logit of beta to logit(0.3)
        init_beta = 0.3
        init_logit_beta = math.log(init_beta / (1 - init_beta))
        self.logit_beta = nn.Parameter(
            torch.tensor([init_logit_beta], dtype=torch.float32)
        )

    def predict(self, representation: Tensor):
        # Apply sigmoid to transform logit_beta back to the range (0, 1)
        beta = torch.sigmoid(self.logit_beta)
        return wiggle(super().predict(representation), beta)


class SingleValueBinarySelectionModel(AbstractBinarySelectionModel):
    """A one parameter selection model as a baseline."""

    def __init__(self, output_dim: int = 1):
        super().__init__()
        self.single_value = nn.Parameter(torch.tensor(0.0))
        self.output_dim = output_dim

    @property
    def hyperparameters(self):
        return {"output_dim": self.output_dim}

    def forward(self, amino_acid_indices: Tensor, mask: Tensor) -> Tensor:
        """Build a binary log selection matrix from an index-encoded parent sequence."""
        if self.output_dim == 1:
            return self.single_value.expand(amino_acid_indices.shape)
        else:
            return self.single_value.expand(
                amino_acid_indices.shape + (self.output_dim,)
            )


class HitClassModel(nn.Module):
    def __init__(self):
        super().__init__()
        self.reinitialize_weights()

    @property
    def hyperparameters(self):
        return {}

    def forward(
        self, parent_codon_idxs: torch.Tensor, uncorrected_codon_probs: torch.Tensor
    ):
        """Forward function takes a tensor of target codon distributions, for each
        observed parent codon, and adjusts the distributions according to the hit class
        corrections."""
        return apply_multihit_correction(
            parent_codon_idxs, uncorrected_codon_probs, self.values
        )

    def reinitialize_weights(self):
        self.values = nn.Parameter(torch.tensor([0.0, 0.0, 0.0]))



import pytest
from netam.framework import (
    load_pcp_df,
    add_shm_model_outputs_to_pcp_df,
)
from netam import pretrained


@pytest.fixture(scope="module")
def pcp_df():
    df = load_pcp_df(
        "data/wyatt-10x-1p5m_pcp_2023-11-30_NI.first100.csv.gz",
    )
    df = add_shm_model_outputs_to_pcp_df(
        df,
        pretrained.load("ThriftyHumV0.2-45"),
    )
    return df



import os

import pandas as pd
import pytest
import numpy as np
import torch

import netam.framework as framework
from netam.common import BIG
from netam.framework import SHMoofDataset, SHMBurrito, RSSHMBurrito
from netam.models import SHMoofModel, RSSHMoofModel, IndepRSCNNModel


@pytest.fixture
def tiny_dataset():
    df = pd.DataFrame({"parent": ["ATGTA", "GTAC"], "child": ["ACGTA", "ATAC"]})
    return SHMoofDataset(df, site_count=6, kmer_length=3)


@pytest.fixture
def tiny_val_dataset():
    df = pd.DataFrame({"parent": ["ATGTA", "GTAA"], "child": ["ACGTA", "TACG"]})
    return SHMoofDataset(df, site_count=6, kmer_length=3)


@pytest.fixture
def tiny_model():
    return SHMoofModel(site_count=6, kmer_length=3)


@pytest.fixture
def tiny_burrito(tiny_dataset, tiny_val_dataset, tiny_model):
    burrito = SHMBurrito(tiny_dataset, tiny_val_dataset, tiny_model)
    burrito.simple_train(epochs=5)
    return burrito


def test_make_dataset(tiny_dataset):
    (
        encoded_parent,
        mask,
        mutation_indicator,
        new_base_idxs,
        wt_base_modifier,
        branch_length,
    ) = tiny_dataset[0]
    assert (mask == torch.tensor([1, 1, 1, 1, 1, 0], dtype=torch.bool)).all()
    # First kmer is NAT due to padding, but our encoding defaults this to "N".
    assert encoded_parent[0].item() == tiny_dataset.encoder.kmer_to_index["N"]
    assert (
        mutation_indicator == torch.tensor([0, 1, 0, 0, 0, 0], dtype=torch.bool)
    ).all()
    assert (
        new_base_idxs == torch.tensor([-1, 1, -1, -1, -1, -1], dtype=torch.int64)
    ).all()
    correct_wt_base_modifier = torch.zeros((6, 4))
    correct_wt_base_modifier[0, 0] = -BIG
    correct_wt_base_modifier[1, 3] = -BIG
    correct_wt_base_modifier[2, 2] = -BIG
    correct_wt_base_modifier[3, 3] = -BIG
    correct_wt_base_modifier[4, 0] = -BIG
    assert (wt_base_modifier == correct_wt_base_modifier).all()
    assert branch_length == 1 / 5


def test_write_tinyburrito_output(tiny_burrito):
    os.makedirs("_ignore", exist_ok=True)
    tiny_burrito.model.write_shmoof_output("_ignore")


def test_crepe_roundtrip(tiny_burrito):
    os.makedirs("_ignore", exist_ok=True)
    tiny_burrito.save_crepe("_ignore/tiny_crepe")
    crepe = framework.load_crepe("_ignore/tiny_crepe")
    assert crepe.encoder.parameters["site_count"] == tiny_burrito.model.site_count
    assert crepe.encoder.parameters["kmer_length"] == tiny_burrito.model.kmer_length
    assert torch.isclose(crepe.model.kmer_rates, tiny_burrito.model.kmer_rates).all()
    assert torch.isclose(crepe.model.site_rates, tiny_burrito.model.site_rates).all()
    ## Assert that crepe.model is in eval mode
    assert not crepe.model.training


@pytest.fixture
def tiny_rsshmoofmodel():
    return RSSHMoofModel(site_count=6, kmer_length=3)


@pytest.fixture
def tiny_rsscnnmodel():
    return IndepRSCNNModel(
        kmer_length=3, embedding_dim=2, filter_count=2, kernel_size=3
    )


@pytest.fixture
def mini_dataset():
    df = pd.read_csv("data/wyatt-10x-1p5m_pcp_2023-11-30_NI.first100.csv.gz")
    return SHMoofDataset(df, site_count=500, kmer_length=3)


@pytest.fixture
def mini_rsburrito(mini_dataset, tiny_rsscnnmodel):
    burrito = RSSHMBurrito(mini_dataset, mini_dataset, tiny_rsscnnmodel)
    burrito.joint_train(epochs=5, training_method="yun")
    return burrito


def test_write_mini_rsburrito_output(mini_rsburrito):
    os.makedirs("_ignore", exist_ok=True)
    mini_rsburrito.save_crepe("_ignore/mini_rscrepe")


def test_standardize_model_rates(mini_rsburrito):
    mini_rsburrito.standardize_model_rates()
    vrc01_rate_14 = mini_rsburrito.vrc01_site_14_model_rate()
    assert np.isclose(vrc01_rate_14, 1.0)



import math
import inspect
import itertools
import resource
import subprocess
from tqdm import tqdm
from functools import wraps
from itertools import islice

import numpy as np
import torch
import torch.optim as optim
from torch import nn, Tensor
import multiprocessing as mp

from netam.sequences import iter_codons, apply_aa_mask_to_nt_sequence

BIG = 1e9
SMALL_PROB = 1e-6
BASES = ["A", "C", "G", "T"]
BASES_AND_N_TO_INDEX = {"A": 0, "C": 1, "G": 2, "T": 3, "N": 4}
AA_STR_SORTED = "ACDEFGHIKLMNPQRSTVWY"
AA_STR_SORTED_AMBIG = AA_STR_SORTED + "X"
MAX_AMBIG_AA_IDX = len(AA_STR_SORTED_AMBIG) - 1

# I needed some sequence to use to normalize the rate of mutation in the SHM model.
# So, I chose perhaps the most famous antibody sequence, VRC01:
# https://www.ncbi.nlm.nih.gov/nuccore/GU980702.1
VRC01_NT_SEQ = (
    "CAGGTGCAGCTGGTGCAGTCTGGGGGTCAGATGAAGAAGCCTGGCGAGTCGATGAGAATT"
    "TCTTGTCGGGCTTCTGGATATGAATTTATTGATTGTACGCTAAATTGGATTCGTCTGGCC"
    "CCCGGAAAAAGGCCTGAGTGGATGGGATGGCTGAAGCCTCGGGGGGGGGCCGTCAACTAC"
    "GCACGTCCACTTCAGGGCAGAGTGACCATGACTCGAGACGTTTATTCCGACACAGCCTTT"
    "TTGGAGCTGCGCTCGTTGACAGTAGACGACACGGCCGTCTACTTTTGTACTAGGGGAAAA"
    "AACTGTGATTACAATTGGGACTTCGAACACTGGGGCCGGGGCACCCCGGTCATCGTCTCA"
    "TCA"
)


def force_spawn():
    """Force the spawn start method for multiprocessing.

    This is necessary to avoid conflicts with the internal OpenMP-based thread pool in
    PyTorch.
    """
    mp.set_start_method("spawn", force=True)


def generate_kmers(kmer_length):
    # Our strategy for kmers is to have a single representation for any kmer that isn't in ACGT.
    # This is the first one, which is simply "N", and so this placeholder value is 0.
    all_kmers = ["N"] + [
        "".join(p) for p in itertools.product(BASES, repeat=kmer_length)
    ]
    assert len(all_kmers) < torch.iinfo(torch.int32).max
    return all_kmers


def kmer_to_index_of(all_kmers):
    return {kmer: idx for idx, kmer in enumerate(all_kmers)}


def aa_idx_tensor_of_str_ambig(aa_str):
    """Return the indices of the amino acids in a string, allowing the ambiguous
    character."""
    try:
        return torch.tensor(
            [AA_STR_SORTED_AMBIG.index(aa) for aa in aa_str], dtype=torch.int
        )
    except ValueError:
        print(f"Found an invalid amino acid in the string: {aa_str}")
        raise


def generic_mask_tensor_of(ambig_symb, seq_str, length=None):
    """Return a mask tensor indicating non-empty and non-ambiguous sites.

    Sites beyond the length of the sequence are masked.
    """
    if length is None:
        length = len(seq_str)
    mask = torch.zeros(length, dtype=torch.bool)
    if len(seq_str) < length:
        seq_str += ambig_symb * (length - len(seq_str))
    else:
        seq_str = seq_str[:length]
    mask[[c != ambig_symb for c in seq_str]] = 1
    return mask


def codon_mask_tensor_of(nt_parent, *other_nt_seqs, aa_length=None):
    """Return a mask tensor indicating codons which contain at least one N.

    Codons beyond the length of the sequence are masked. If other_nt_seqs are provided,
    the "and" mask will be computed for all sequences
    """
    if aa_length is None:
        aa_length = len(nt_parent) // 3
    sequences = (nt_parent,) + other_nt_seqs
    mask = [
        all("N" not in codon for codon in codons)
        for codons in zip(*(iter_codons(sequence) for sequence in sequences))
    ]
    if len(mask) < aa_length:
        mask += [False] * (aa_length - len(mask))
    else:
        mask = mask[:aa_length]
    assert len(mask) == aa_length
    return torch.tensor(mask, dtype=torch.bool)


def assert_pcp_valid(parent, child, aa_mask=None):
    """Check that the parent-child pairs are valid.

    * The parent and child sequences must be the same length
    * There must be unmasked codons
    * The parent and child sequences must not match after masking codons containing
      ambiguities.

    Args:
        parent: The parent sequence.
        child: The child sequence.
        aa_mask: The mask tensor for the amino acid sequence. If None, it will be
            computed from the parent and child sequences.
    """
    if aa_mask is None:
        aa_mask = codon_mask_tensor_of(parent, child)
    if len(parent) != len(child):
        raise ValueError("Parent and child sequences are not the same length.")
    if not aa_mask.any():
        raise ValueError("Parent-child pair is masked in all codons.")
    if apply_aa_mask_to_nt_sequence(parent, aa_mask) == apply_aa_mask_to_nt_sequence(
        child, aa_mask
    ):
        raise ValueError(
            "Parent-child pair matches after masking codons containing ambiguities"
        )


def nt_mask_tensor_of(*args, **kwargs):
    return generic_mask_tensor_of("N", *args, **kwargs)


def aa_mask_tensor_of(*args, **kwargs):
    return generic_mask_tensor_of("X", *args, **kwargs)


def informative_site_count(seq_str):
    return sum(c != "N" for c in seq_str)


def clamp_probability(x: Tensor) -> Tensor:
    return torch.clamp(x, min=SMALL_PROB, max=(1.0 - SMALL_PROB))


def clamp_log_probability(x: Tensor) -> Tensor:
    return torch.clamp(x, max=np.log(1.0 - SMALL_PROB))


def print_parameter_count(model):
    total = 0
    for name, module in model.named_modules():
        if len(list(module.children())) == 0:  # Only count parameters in leaf modules
            num_params = sum(p.numel() for p in module.parameters())
            print(f"{name}: {num_params} parameters")
            total += num_params
    print("-----")
    print(f"total: {total} parameters")


def parameter_count_of_model(model):
    return sum(p.numel() for p in model.parameters() if p.requires_grad)


def stack_heterogeneous(tensors, pad_value=0.0):
    """Stack an iterable of 1D or 2D torch.Tensor objects of different lengths along the
    first dimension into a single tensor.

        black --check netam tests
    Args:
        tensors (iterable): An iterable of 1D or 2D torch.Tensor objects with variable lengths in the first dimension.
        pad_value (number): The value used for padding shorter tensors. Default is 0.

    Returns:
        torch.Tensor: A stacked tensor with all input tensors padded to the length of the longest tensor in the first dimension.
    """
    if tensors is None or len(tensors) == 0:
        return torch.Tensor()  # Return an empty tensor if no tensors are provided

    dim = tensors[0].dim()
    if dim not in [1, 2]:
        raise ValueError("This function only supports 1D or 2D tensors.")

    max_length = max(tensor.size(0) for tensor in tensors)

    if dim == 1:
        # If 1D, simply pad the end of the tensor.
        padded_tensors = [
            torch.nn.functional.pad(
                tensor, (0, max_length - tensor.size(0)), value=pad_value
            )
            for tensor in tensors
        ]
    else:
        # If 2D, pad the end of the first dimension (rows); the argument to pad
        # is a tuple of (padding_left, padding_right, padding_top,
        # padding_bottom)
        padded_tensors = [
            torch.nn.functional.pad(
                tensor, (0, 0, 0, max_length - tensor.size(0)), value=pad_value
            )
            for tensor in tensors
        ]

    return torch.stack(padded_tensors)


def optimizer_of_name(optimizer_name, model_parameters, **kwargs):
    """Build a torch.optim optimizer from a string name and model parameters.

    Use a SGD optimizer with momentum if the optimizer_name is "SGDMomentum".
    """
    if optimizer_name == "SGDMomentum":
        optimizer_name = "SGD"
        kwargs["momentum"] = 0.9
    try:
        optimizer_class = getattr(optim, optimizer_name)
        return optimizer_class(model_parameters, **kwargs)
    except AttributeError:
        raise ValueError(
            f"Optimizer '{optimizer_name}' is not recognized in torch.optim"
        )


def find_least_used_cuda_gpu():
    """Find the least used CUDA GPU on the system using nvidia-smi.

    If they are all idle, return None.
    """
    result = subprocess.run(
        ["nvidia-smi", "--query-gpu=utilization.gpu", "--format=csv,nounits,noheader"],
        stdout=subprocess.PIPE,
        text=True,
    )
    if result.returncode != 0:
        print(f"Error running nvidia-smi.")
        return None
    utilization = [int(x) for x in result.stdout.strip().split("\n")]
    if max(utilization) == 0:
        return None  # All GPUs are idle.
    # else:
    return utilization.index(min(utilization))


def pick_device(gpu_preference=None):
    """Pick a device for PyTorch to use.

    If gpu_preference is a string, use the device with that name. This is considered a
    strong preference from a user who knows what they are doing.

    If gpu_preference is an integer, this is a weak preference for a numbered GPU.  If
    CUDA is available, use the least used GPU, and if all are idle use the gpu_index
    modulo the number of GPUs. If gpu_index is None, then use a random GPU.
    """

    # Strong preference for a specific device.
    if gpu_preference is not None and isinstance(gpu_preference, str):
        return torch.device(gpu_preference)

    # else weak preference for a numbered GPU.

    # check that CUDA is usable
    def check_CUDA():
        try:
            torch._C._cuda_init()
            return True
        except:
            return False

    if torch.backends.cudnn.is_available() and check_CUDA():
        which_gpu = find_least_used_cuda_gpu()
        if which_gpu is None:
            if gpu_preference is None:
                which_gpu = np.random.randint(torch.cuda.device_count())
            else:
                which_gpu = gpu_preference % torch.cuda.device_count()
        print(f"Using CUDA GPU {which_gpu}")
        return torch.device(f"cuda:{which_gpu}")
    elif torch.backends.mps.is_available():
        print("Using Metal Performance Shaders")
        return torch.device("mps")
    else:
        print("Using CPU")
        return torch.device("cpu")


def print_tensor_devices(scope="local"):
    """Print the devices of all PyTorch tensors in the given scope.

    Args:
        scope (str): 'local' for local scope, 'global' for global scope.
    """
    if scope == "local":
        frame = inspect.currentframe()
        variables = frame.f_back.f_locals
    elif scope == "global":
        variables = globals()
    else:
        raise ValueError("Scope must be 'local' or 'global'.")

    for var_name, var_value in variables.items():
        if isinstance(var_value, torch.Tensor):
            print(f"{var_name}: {var_value.device}")


def get_memory_usage_mb():
    # Returns the peak memory usage in MB
    usage = resource.getrusage(resource.RUSAGE_SELF)
    return usage.ru_maxrss / 1024  # Convert from KB to MB


def tensor_to_np_if_needed(x):
    if isinstance(x, torch.Tensor):
        return x.cpu().numpy()
    else:
        assert isinstance(x, np.ndarray)
        return x


# Reference: https://pytorch.org/tutorials/beginner/transformer_tutorial.html
class PositionalEncoding(nn.Module):
    def __init__(self, d_model: int, dropout: float = 0.1, max_len: int = 5000):
        super().__init__()
        self.dropout = nn.Dropout(p=dropout)

        # assert that d_model is even
        assert d_model % 2 == 0, "d_model must be even for PositionalEncoding"

        position = torch.arange(max_len).unsqueeze(1)
        div_term = torch.exp(
            torch.arange(0, d_model, 2) * (-math.log(10000.0) / d_model)
        )
        pe = torch.zeros(max_len, 1, d_model)
        pe[:, 0, 0::2] = torch.sin(position * div_term)
        pe[:, 0, 1::2] = torch.cos(position * div_term)
        self.register_buffer("pe", pe)

    def forward(self, x: Tensor) -> Tensor:
        """
        Arguments:
            x: Tensor, shape ``[seq_len, batch_size, embedding_dim]``
        """
        x = x + self.pe[: x.size(0)]
        return self.dropout(x)


def linear_bump_lr(epoch, warmup_epochs, total_epochs, max_lr, min_lr):
    """Linearly increase the learning rate from min_lr to max_lr over warmup_epochs,
    then linearly decrease the learning rate from max_lr to min_lr.

    See https://github.com/matsengrp/netam/pull/41 for more details.

    Example:
    .. code-block:: python
        pd.Series([linear_bump_lr(epoch, warmup_epochs=20, total_epochs=200, max_lr=0.01, min_lr=1e-5) for epoch in range(200)]).plot()
    """
    if epoch < warmup_epochs:
        lr = min_lr + ((max_lr - min_lr) / warmup_epochs) * epoch
    else:
        lr = max_lr - ((max_lr - min_lr) / (total_epochs - warmup_epochs)) * (
            epoch - warmup_epochs
        )
    return lr


def encode_sequences(sequences, encoder):
    encoded_parents, wt_base_modifiers = zip(
        *[encoder.encode_sequence(sequence) for sequence in sequences]
    )
    masks = [nt_mask_tensor_of(sequence, encoder.site_count) for sequence in sequences]
    return (
        torch.stack(encoded_parents),
        torch.stack(masks),
        torch.stack(wt_base_modifiers),
    )


# from https://docs.python.org/3.11/library/itertools.html#itertools-recipes
# avoiding walrus:
def chunked(iterable, n):
    "Chunk data into lists of length n. The last chunk may be shorter."
    it = iter(iterable)
    while True:
        chunk = list(islice(it, n))
        if not chunk:
            return
        yield chunk


def chunk_method(default_chunk_size=2048, progress_bar_name=None):
    """Decorator to chunk the input to a method.

    Expects that all positional arguments are iterables of the same length,
    and that outputs are tuples of tensors whose first dimension
    corresponds to the first dimension of the input iterables.

    If method returns just one item, it must not be a tuple.

    Chunking is done along the first dimension of all inputs.

    Args:
        default_chunk_size: The default chunk size. The decorated method can
            also automatically accept a `default_chunk_size` keyword argument.
        progress_bar_name: The name of the progress bar. If None, no progress bar is shown.
    """

    def decorator(method):
        @wraps(method)
        def wrapper(self, *args, **kwargs):
            if "chunk_size" in kwargs:
                chunk_size = kwargs.pop("chunk_size")
            else:
                chunk_size = default_chunk_size
            results = []
            if progress_bar_name is None:
                progargs = {"disable": True}
            else:
                progargs = {"desc": progress_bar_name}
            bar = tqdm(total=len(args[0]), delay=2.0, **progargs)
            for chunked_args in zip(*(chunked(arg, chunk_size) for arg in args)):
                bar.update(len(chunked_args[0]))
                results.append(method(self, *chunked_args, **kwargs))
            if isinstance(results[0], tuple):
                return tuple(torch.cat(tensors) for tensors in zip(*results))
            else:
                return torch.cat(results)

        return wrapper

    return decorator



"""This module provides a simple interface for downloading pre-trained models.

It was inspired by the `load_model` module of [AbLang2](https://github.com/oxpig/AbLang2).
"""

import os
import zipfile
from importlib.resources import files

import requests

from netam.framework import load_crepe

with files(__package__).joinpath("_pretrained") as pretrained_path:
    PRETRAINED_DIR = str(pretrained_path)

PACKAGE_LOCATIONS_AND_CONTENTS = (
    # Order of entries:
    # * Local file name
    # * Remote URL
    # * Directory in which the models appear after extraction (must match path determined by archive)
    # * List of models in the package
    [
        "thrifty-0.2.0.zip",
        "https://github.com/matsengrp/thrifty-models/archive/refs/tags/v0.2.0.zip",
        "thrifty-models-0.2.0/models",
        [
            "ThriftyHumV0.2-20",
            "ThriftyHumV0.2-45",
            "ThriftyHumV0.2-59",
        ],
    ],
)

LOCAL_TO_REMOTE = {}
MODEL_TO_LOCAL = {}
LOCAL_TO_DIR = {}

for local_file, remote, models_dir, models in PACKAGE_LOCATIONS_AND_CONTENTS:
    LOCAL_TO_REMOTE[local_file] = remote

    for model in models:
        MODEL_TO_LOCAL[model] = (local_file, models_dir)


def local_path_for_model(model_name: str):
    """Return the local path for a model, downloading it if necessary."""

    if model_name not in MODEL_TO_LOCAL:
        raise ValueError(f"Model {model_name} not found in pre-trained models.")

    os.makedirs(PRETRAINED_DIR, exist_ok=True)

    local_package, models_dir = MODEL_TO_LOCAL[model_name]
    local_package_path = os.path.join(PRETRAINED_DIR, local_package)

    if not os.path.exists(local_package_path):
        url = LOCAL_TO_REMOTE[local_package]
        print(f"Fetching models: downloading {url} to {local_package_path}")
        response = requests.get(url)
        response.raise_for_status()
        with open(local_package_path, "wb") as f:
            f.write(response.content)
        if local_package.endswith(".zip"):
            with zipfile.ZipFile(local_package_path, "r") as zip_ref:
                zip_ref.extractall(path=PRETRAINED_DIR)
        else:
            raise ValueError(f"Unknown file type for {local_package}")
    else:
        print(f"Using cached models: {local_package_path}")

    local_crepe_path = os.path.join(PRETRAINED_DIR, models_dir, model_name)

    if not os.path.exists(local_crepe_path + ".yml"):
        raise ValueError(f"Model {local_crepe_path} not found in pre-trained models.")
    if not os.path.exists(local_crepe_path + ".pth"):
        raise ValueError(f"Model {model_name} missing model weights.")

    return local_crepe_path


def load(model_name: str, device=None):
    """Load a pre-trained model.

    If the model is not already downloaded, it will be downloaded from the appropriate
    URL and stashed in the PRETRAINED_DIR.
    """

    local_crepe_path = local_path_for_model(model_name)
    return load_crepe(local_crepe_path, device=device)



import netam.pretrained as pretrained
from pathlib import Path
import shutil


def test_names_unique():
    # Check all defined models can be loaded and they have unique names
    assert len(set(pretrained.MODEL_TO_LOCAL.keys())) == sum(
        len(models) for _, _, _, models in pretrained.PACKAGE_LOCATIONS_AND_CONTENTS
    )


def test_load_all_models():
    # Remove cached models:
    shutil.rmtree(Path(pretrained.PRETRAINED_DIR))
    # Check each can be loaded without caching
    for model_name in pretrained.MODEL_TO_LOCAL:
        pretrained.load(model_name)

    # Check each can be loaded with caching
    for model_name in pretrained.MODEL_TO_LOCAL:
        pretrained.load(model_name)



import torch
import pytest

import netam.molevol as molevol
from netam import pretrained

from netam.sequences import (
    nt_idx_tensor_of_str,
    translate_sequence,
    AA_STR_SORTED,
    CODONS,
    NT_STR_SORTED,
)

# These happen to be the same as some examples in test_models.py but that's fine.
# If it was important that they were shared, we should put them in a conftest.py.
ex_mut_probs = torch.tensor([0.01, 0.02, 0.03])
ex_csps = torch.tensor(
    [[0.0, 0.3, 0.5, 0.2], [0.4, 0.0, 0.1, 0.5], [0.2, 0.3, 0.0, 0.5]]
)

ex_parent_codon_idxs = nt_idx_tensor_of_str("ACG")
parent_nt_seq = "CAGGTGCAGCTGGTGGAG"  # QVQLVE


def test_aaprobs_of_parent_scaled_rates_and_csps():

    def old_aaprobs_of_parent_scaled_rates_and_csps(
        parent_idxs: torch.Tensor, scaled_rates: torch.Tensor, csps: torch.Tensor
    ) -> torch.Tensor:
        """Calculate per-site amino acid probabilities from per-site nucleotide rates
        and substitution probabilities.

        Args:
            parent_idxs (torch.Tensor): Parent nucleotide indices. Shape should be (site_count,).
            scaled_rates (torch.Tensor): Poisson rates of mutation per site, scaled by branch length.
                                         Shape should be (site_count,).
            csps (torch.Tensor): Substitution probabilities per site: a 2D
                                      tensor with shape (site_count, 4).

        Returns:
            torch.Tensor: A 2D tensor with rows corresponding to sites and columns
                          corresponding to amino acids.
        """
        # Calculate the probability of at least one mutation at each site.
        mut_probs = 1.0 - torch.exp(-scaled_rates)

        # Reshape the inputs to include a codon dimension.
        parent_codon_idxs = molevol.reshape_for_codons(parent_idxs)
        codon_mut_probs = molevol.reshape_for_codons(mut_probs)
        codon_csps = molevol.reshape_for_codons(csps)

        # Vectorized calculation of amino acid probabilities.
        return molevol.aaprob_of_mut_and_sub(
            parent_codon_idxs, codon_mut_probs, codon_csps
        )

    assert torch.allclose(
        old_aaprobs_of_parent_scaled_rates_and_csps(
            ex_parent_codon_idxs, ex_mut_probs, ex_csps
        ),
        molevol.aaprobs_of_parent_scaled_rates_and_csps(
            ex_parent_codon_idxs, ex_mut_probs, ex_csps
        ),
    )


def test_build_mutation_matrix():
    correct_tensor = torch.tensor(
        [
            # probability of mutation to each nucleotide (first entry in the first row
            # is probability of no mutation)
            [0.99, 0.003, 0.005, 0.002],
            [0.008, 0.98, 0.002, 0.01],
            [0.006, 0.009, 0.97, 0.015],
        ]
    )

    computed_tensor = molevol.build_mutation_matrices(
        ex_parent_codon_idxs.unsqueeze(0),
        ex_mut_probs.unsqueeze(0),
        ex_csps.unsqueeze(0),
    ).squeeze()

    assert torch.allclose(correct_tensor, computed_tensor)


def test_neutral_aa_mut_probs():
    # This is the probability of a mutation to a codon that translates to the
    # same. In this case, ACG is the codon, and it's fourfold degenerate. Thus
    # we just multiply the probability of A and C staying the same from the
    # correct_tensor just above.
    correct_tensor = torch.tensor([1 - 0.99 * 0.98])

    computed_tensor = molevol.neutral_aa_mut_probs(
        ex_parent_codon_idxs.unsqueeze(0),
        ex_mut_probs.unsqueeze(0),
        ex_csps.unsqueeze(0),
    ).squeeze()

    assert torch.allclose(correct_tensor, computed_tensor)


def test_check_csps():
    parent_idxs = nt_idx_tensor_of_str("AC")
    csp = torch.tensor([[0.0, 0.375, 0.5, 0.125], [0.125, 0.0, 0.375, 0.5]])
    molevol.check_csps(parent_idxs, csp)

    not_csp = torch.tensor([[0.2, 0.3, 0.4, 0.1], [0.1, 0.2, 0.3, 0.4]])
    with pytest.raises(AssertionError):
        molevol.check_csps(parent_idxs, not_csp)


def iterative_aaprob_of_mut_and_sub(parent_codon, mut_probs, csps):
    """Original version of codon_to_aa_probabilities, used for testing."""
    aa_probs = {}
    for aa in AA_STR_SORTED:
        aa_probs[aa] = 0.0

    # iterate through all possible child codons
    for child_codon in CODONS:
        try:
            aa = translate_sequence(child_codon)
        except ValueError:  # check for STOP codon
            continue

        # iterate through codon sites and compute total probability of potential child codon
        child_prob = 1.0
        for isite in range(3):
            if parent_codon[isite] == child_codon[isite]:
                child_prob *= 1.0 - mut_probs[isite]
            else:
                child_prob *= mut_probs[isite]
                child_prob *= csps[isite][NT_STR_SORTED.index(child_codon[isite])]

        aa_probs[aa] += child_prob

    # need renormalization factor so that amino acid probabilities sum to 1,
    # since probabilities to STOP codon are dropped
    psum = sum(aa_probs.values())

    return torch.tensor([aa_probs[aa] / psum for aa in AA_STR_SORTED])


def test_aaprob_of_mut_and_sub():
    crepe = pretrained.load("ThriftyHumV0.2-45")
    [rates], [subs] = crepe([parent_nt_seq])
    mut_probs = 1.0 - torch.exp(-rates.squeeze().clone().detach())
    parent_codon = parent_nt_seq[0:3]
    parent_codon_idxs = nt_idx_tensor_of_str(parent_codon)
    codon_mut_probs = mut_probs[0:3]
    codon_subs = subs.clone().detach()[0:3]

    iterative_result = iterative_aaprob_of_mut_and_sub(
        parent_codon, codon_mut_probs, codon_subs
    )

    parent_codon_idxs = parent_codon_idxs.unsqueeze(0)
    codon_mut_probs = codon_mut_probs.unsqueeze(0)
    codon_subs = codon_subs.unsqueeze(0)

    assert torch.allclose(
        iterative_result,
        molevol.aaprob_of_mut_and_sub(
            parent_codon_idxs,
            codon_mut_probs,
            codon_subs,
        ).squeeze(),
    )



import torch

from netam.common import nt_mask_tensor_of, aa_mask_tensor_of, codon_mask_tensor_of


def test_mask_tensor_of():
    input_seq = "NAAA"
    # First test as nucleotides.
    expected_output = torch.tensor([0, 1, 1, 1, 0], dtype=torch.bool)
    output = nt_mask_tensor_of(input_seq, length=5)
    assert torch.equal(output, expected_output)
    # Next test as amino acids, where N counts as an AA.
    expected_output = torch.tensor([1, 1, 1, 1, 0], dtype=torch.bool)
    output = aa_mask_tensor_of(input_seq, length=5)
    assert torch.equal(output, expected_output)


def test_codon_mask_tensor_of():
    input_seq = "NAAAAAAAAAA"
    # First test as nucleotides.
    expected_output = torch.tensor([0, 1, 1, 0, 0], dtype=torch.bool)
    output = codon_mask_tensor_of(input_seq, aa_length=5)
    assert torch.equal(output, expected_output)
    input_seq2 = "AAAANAAAAAA"
    expected_output = torch.tensor([0, 0, 1, 0, 0], dtype=torch.bool)
    output = codon_mask_tensor_of(input_seq, input_seq2, aa_length=5)
    assert torch.equal(output, expected_output)



"""Here we define a model that outputs a vector of 20 amino acid preferences."""

import torch
import torch.nn.functional as F

from netam.common import (
    clamp_probability,
    BIG,
)
from netam.dxsm import DXSMDataset, DXSMBurrito
import netam.framework as framework
import netam.molevol as molevol
import netam.sequences as sequences
import copy


class DASMDataset(DXSMDataset):
    prefix = "dasm"

    def update_neutral_probs(self):
        neutral_aa_probs_l = []

        for nt_parent, mask, nt_rates, nt_csps, branch_length in zip(
            self.nt_parents,
            self.masks,
            self.nt_ratess,
            self.nt_cspss,
            self._branch_lengths,
        ):
            mask = mask.to("cpu")
            nt_rates = nt_rates.to("cpu")
            nt_csps = nt_csps.to("cpu")
            if self.multihit_model is not None:
                multihit_model = copy.deepcopy(self.multihit_model).to("cpu")
            else:
                multihit_model = None
            # Note we are replacing all Ns with As, which means that we need to be careful
            # with masking out these positions later. We do this below.
            parent_idxs = sequences.nt_idx_tensor_of_str(nt_parent.replace("N", "A"))
            parent_len = len(nt_parent)

            mut_probs = 1.0 - torch.exp(-branch_length * nt_rates[:parent_len])
            nt_csps = nt_csps[:parent_len, :]
            nt_mask = mask.repeat_interleave(3)[: len(nt_parent)]
            molevol.check_csps(parent_idxs[nt_mask], nt_csps[: len(nt_parent)][nt_mask])

            neutral_aa_probs = molevol.neutral_aa_probs(
                parent_idxs.reshape(-1, 3),
                mut_probs.reshape(-1, 3),
                nt_csps.reshape(-1, 3, 4),
                multihit_model=multihit_model,
            )

            if not torch.isfinite(neutral_aa_probs).all():
                print(f"Found a non-finite neutral_aa_probs")
                print(f"nt_parent: {nt_parent}")
                print(f"mask: {mask}")
                print(f"nt_rates: {nt_rates}")
                print(f"nt_csps: {nt_csps}")
                print(f"branch_length: {branch_length}")
                raise ValueError(f"neutral_aa_probs is not finite: {neutral_aa_probs}")

            # Ensure that all values are positive before taking the log later
            neutral_aa_probs = clamp_probability(neutral_aa_probs)

            pad_len = self.max_aa_seq_len - neutral_aa_probs.shape[0]
            if pad_len > 0:
                neutral_aa_probs = F.pad(
                    neutral_aa_probs, (0, 0, 0, pad_len), value=1e-8
                )
            # Here we zero out masked positions.
            neutral_aa_probs *= mask[:, None]

            neutral_aa_probs_l.append(neutral_aa_probs)

        # Note that our masked out positions will have a nan log probability,
        # which will require us to handle them correctly downstream.
        self.log_neutral_aa_probss = torch.log(torch.stack(neutral_aa_probs_l))

    def __getitem__(self, idx):
        return {
            "aa_parents_idxs": self.aa_parents_idxss[idx],
            "aa_children_idxs": self.aa_children_idxss[idx],
            "subs_indicator": self.aa_subs_indicators[idx],
            "mask": self.masks[idx],
            "log_neutral_aa_probs": self.log_neutral_aa_probss[idx],
            "nt_rates": self.nt_ratess[idx],
            "nt_csps": self.nt_cspss[idx],
        }

    def to(self, device):
        self.aa_parents_idxss = self.aa_parents_idxss.to(device)
        self.aa_children_idxss = self.aa_children_idxss.to(device)
        self.aa_subs_indicators = self.aa_subs_indicators.to(device)
        self.masks = self.masks.to(device)
        self.log_neutral_aa_probss = self.log_neutral_aa_probss.to(device)
        self.nt_ratess = self.nt_ratess.to(device)
        self.nt_cspss = self.nt_cspss.to(device)
        if self.multihit_model is not None:
            self.multihit_model = self.multihit_model.to(device)


def zap_predictions_along_diagonal(predictions, aa_parents_idxs):
    """Set the diagonal (i.e. no amino acid change) of the predictions tensor to -BIG,
    except where aa_parents_idxs >= 20, which indicates no update should be done."""

    device = predictions.device
    batch_size, L, _ = predictions.shape
    batch_indices = torch.arange(batch_size, device=device)[:, None].expand(-1, L)
    sequence_indices = torch.arange(L, device=device)[None, :].expand(batch_size, -1)

    # Create a mask for valid positions (where aa_parents_idxs is less than 20)
    valid_mask = aa_parents_idxs < 20

    # Only update the predictions for valid positions
    predictions[
        batch_indices[valid_mask],
        sequence_indices[valid_mask],
        aa_parents_idxs[valid_mask],
    ] = -BIG

    return predictions


class DASMBurrito(framework.TwoLossMixin, DXSMBurrito):
    prefix = "dasm"

    def __init__(self, *args, loss_weights: list = [1.0, 0.01], **kwargs):
        super().__init__(*args, **kwargs)
        self.xent_loss = torch.nn.CrossEntropyLoss()
        self.loss_weights = torch.tensor(loss_weights).to(self.device)

    def prediction_pair_of_batch(self, batch):
        """Get log neutral AA probabilities and log selection factors for a batch of
        data."""
        aa_parents_idxs = batch["aa_parents_idxs"].to(self.device)
        mask = batch["mask"].to(self.device)
        log_neutral_aa_probs = batch["log_neutral_aa_probs"].to(self.device)
        if not torch.isfinite(log_neutral_aa_probs[mask]).all():
            raise ValueError(
                f"log_neutral_aa_probs has non-finite values at relevant positions: {log_neutral_aa_probs[mask]}"
            )
        log_selection_factors = self.model(aa_parents_idxs, mask)
        return log_neutral_aa_probs, log_selection_factors

    def predictions_of_pair(self, log_neutral_aa_probs, log_selection_factors):
        """Take the sum of the neutral mutation log probabilities and the selection
        factors.

        In contrast to a DNSM, each of these now have last dimension of 20.
        """
        predictions = log_neutral_aa_probs + log_selection_factors
        assert torch.isnan(predictions).sum() == 0
        return predictions

    def predictions_of_batch(self, batch):
        """Make predictions for a batch of data.

        Note that we use the mask for prediction as part of the input for the
        transformer, though we don't mask the predictions themselves.
        """
        log_neutral_aa_probs, log_selection_factors = self.prediction_pair_of_batch(
            batch
        )
        return self.predictions_of_pair(log_neutral_aa_probs, log_selection_factors)

    def loss_of_batch(self, batch):
        aa_subs_indicator = batch["subs_indicator"].to(self.device)
        # Netam issue #16: child mask would be preferable here.
        mask = batch["mask"].to(self.device)
        aa_parents_idxs = batch["aa_parents_idxs"].to(self.device)
        aa_children_idxs = batch["aa_children_idxs"].to(self.device)
        masked_aa_subs_indicator = aa_subs_indicator.masked_select(mask)
        predictions = self.predictions_of_batch(batch)

        # "Zapping" out the diagonal means setting it to zero in log space by
        # setting it to -BIG. This is a no-op for sites that have an X
        # (ambiguous AA) in the parent. This could cause problems in principle,
        # but in practice we mask out sites with Xs in the parent for the
        # mut_pos_loss, and we mask out sites with no substitution for the CSP
        # loss. The latter class of sites also eliminates sites that have Xs in
        # the parent or child (see sequences.aa_subs_indicator_tensor_of).
        predictions = zap_predictions_along_diagonal(predictions, aa_parents_idxs)

        # After zapping out the diagonal, we can effectively sum over the
        # off-diagonal elements to get the probability of a nonsynonymous
        # substitution.
        subs_pos_pred = torch.sum(torch.exp(predictions), dim=-1)
        subs_pos_pred = subs_pos_pred.masked_select(mask)
        subs_pos_pred = clamp_probability(subs_pos_pred)
        subs_pos_loss = self.bce_loss(subs_pos_pred, masked_aa_subs_indicator)

        # We now need to calculate the conditional substitution probability
        # (CSP) loss. We have already zapped out the diagonal, and we're in
        # logit space, so we are set up for using the cross entropy loss.
        # However we have to mask out the sites that are not substituted, i.e.
        # the sites for which aa_subs_indicator is 0.
        subs_mask = aa_subs_indicator == 1
        csp_pred = predictions[subs_mask]
        csp_targets = aa_children_idxs[subs_mask]
        csp_loss = self.xent_loss(csp_pred, csp_targets)

        return torch.stack([subs_pos_loss, csp_loss])

    def build_selection_matrix_from_parent(self, parent: str):
        """Build a selection matrix from a parent amino acid sequence.

        Values at ambiguous sites are meaningless.
        """
        # This is simpler than the equivalent in dnsm.py because we get the selection
        # matrix directly. Note that selection_factors_of_aa_str does the exponentiation
        # so this indeed gives us the selection factors, not the log selection factors.
        parent = sequences.translate_sequence(parent)
        per_aa_selection_factors = self.model.selection_factors_of_aa_str(parent)

        parent = parent.replace("X", "A")
        parent_idxs = sequences.aa_idx_array_of_str(parent)
        per_aa_selection_factors[torch.arange(len(parent_idxs)), parent_idxs] = 1.0

        return per_aa_selection_factors



import torch

from netam.framework import encode_mut_pos_and_base


def test_encode_mut_pos_and_base():
    parent = "ACGTACTG"
    child_ = "AGGTACCG"
    site_count = 9

    expected_mutation_indi = torch.tensor([0, 1, 0, 0, 0, 0, 1, 0, 0], dtype=torch.bool)
    expected_new_base_idxs = torch.tensor(
        [-1, 2, -1, -1, -1, -1, 1, -1, -1], dtype=torch.int64
    )

    mutation_indicator, new_base_idxs = encode_mut_pos_and_base(
        parent, child_, site_count
    )

    assert torch.equal(
        mutation_indicator, expected_mutation_indi
    ), "Mutation indicators do not match."
    assert torch.equal(
        new_base_idxs, expected_new_base_idxs
    ), "New base indices do not match."



"""Burrito and Dataset classes for training a model to predict simple hit class
corrections to codon probabilities.

Each codon mutation is hit class 0, 1, 2, or 3, corresponding to 0, 1, 2, or 3 mutations
in the codon.

The hit class corrections are three scalar values, one for each nonzero hit class. To
apply the correction to existing codon probability predictions, we multiply the
probability of each child codon by the correction factor for its hit class, then
renormalize. The correction factor for hit class 0 is fixed at 1.
"""

import torch
from torch.utils.data import Dataset
from tqdm import tqdm
import pandas as pd
from typing import Sequence, List, Tuple

from netam.molevol import (
    reshape_for_codons,
    optimize_branch_length,
    codon_probs_of_parent_scaled_nt_rates_and_csps,
)
from netam.hit_class import hit_class_probs_tensor
from netam import sequences
from netam.common import stack_heterogeneous
import netam.framework as framework
from netam.framework import Burrito
from netam.models import HitClassModel


def _trim_to_codon_boundary_and_max_len(
    seqs: List[Sequence], max_len: int = None
) -> List[Sequence]:
    """Trims sequences to codon boundary and maximum length.

    No assumption is made about the data of a sequence, other than that it is
    indexable (string or list of nucleotide indices both work).

    `max_len` is the maximum number of nucleotides to be preserved.
    If `max_len` is None, does not enforce a maximum length.
    """
    if max_len is None:
        return [seq[: len(seq) - len(seq) % 3] for seq in seqs]
    else:
        max_codon_len = max_len - max_len % 3
        return [seq[: min(len(seq) - len(seq) % 3, max_codon_len)] for seq in seqs]


def _observed_hit_classes(parents: Sequence[str], children: Sequence[str]):
    """Compute the observed hit classes between parent and child sequences.

    Args:
        parents (Sequence[str]): A list of parent sequences.
        children (Sequence[str]): A list of the corresponding child sequences.

    Returns:
        List[torch.Tensor]: A list of tensors, each containing the observed
            hit classes for each codon in the parent sequence. At any codon position
            where the parent or child sequence contains an N, the corresponding tensor
            element will be -100.
    """
    labels = []
    for parent_seq, child_seq in zip(parents, children):

        assert len(parent_seq) == len(child_seq)
        codon_count = len(parent_seq) // 3
        valid_length = codon_count * 3

        # Chunk into codons and count mutations
        num_mutations = []
        for i in range(0, valid_length, 3):
            parent_codon = parent_seq[i : i + 3]
            child_codon = child_seq[i : i + 3]

            if "N" in parent_codon or "N" in child_codon:
                num_mutations.append(-100)
            else:
                # Count differing bases
                mutations = sum(1 for p, c in zip(parent_codon, child_codon) if p != c)
                num_mutations.append(mutations)

        # Pad or truncate the mutation counts to match codon_count
        padded_mutations = num_mutations[:codon_count]  # Truncate if necessary
        padded_mutations += [-100] * (
            codon_count - len(padded_mutations)
        )  # Pad with -100s

        # Update the labels tensor for this row
        labels.append(torch.tensor(padded_mutations, dtype=torch.int))
    return labels


class HitClassDataset(Dataset):
    def __init__(
        self,
        nt_parents: Sequence[str],
        nt_children: Sequence[str],
        nt_ratess: Sequence[List[float]],
        nt_cspss: Sequence[List[List[float]]],
        branch_length_multiplier: float = 1.0,
    ):
        trimmed_parents = _trim_to_codon_boundary_and_max_len(nt_parents)
        trimmed_children = _trim_to_codon_boundary_and_max_len(nt_children)
        self.nt_parents = stack_heterogeneous(
            pd.Series(
                sequences.nt_idx_tensor_of_str(parent.replace("N", "A"))
                for parent in trimmed_parents
            )
        )
        self.nt_children = stack_heterogeneous(
            pd.Series(
                sequences.nt_idx_tensor_of_str(child.replace("N", "A"))
                for child in trimmed_children
            )
        )
        self.nt_ratess = stack_heterogeneous(
            pd.Series(_trim_to_codon_boundary_and_max_len(nt_ratess)).reset_index(
                drop=True
            )
        )
        self.nt_cspss = stack_heterogeneous(
            pd.Series(_trim_to_codon_boundary_and_max_len(nt_cspss)).reset_index(
                drop=True
            )
        )

        assert len(self.nt_parents) == len(self.nt_children)

        for parent, child in zip(trimmed_parents, trimmed_children):
            if parent == child:
                raise ValueError(
                    f"Found an identical parent and child sequence: {parent}"
                )
            assert len(parent) == len(child)

        self.observed_hcs = stack_heterogeneous(
            _observed_hit_classes(trimmed_parents, trimmed_children), pad_value=-100
        ).long()

        # This masks any codon that has an N in the parent or child. This may
        # change with https://github.com/matsengrp/netam/issues/16
        self.codon_mask = self.observed_hcs > -1

        # Make initial branch lengths (will get optimized later).
        self._branch_lengths = torch.tensor(
            [
                sequences.nt_mutation_frequency(parent, child)
                * branch_length_multiplier
                for parent, child in zip(trimmed_parents, trimmed_children)
            ]
        )
        self.update_hit_class_probs()

    @property
    def branch_lengths(self):
        return self._branch_lengths

    @branch_lengths.setter
    def branch_lengths(self, new_branch_lengths):
        assert len(new_branch_lengths) == len(self._branch_lengths), (
            f"Expected {len(self._branch_lengths)} branch lengths, "
            f"got {len(new_branch_lengths)}"
        )
        assert (torch.isfinite(new_branch_lengths) & (new_branch_lengths > 0)).all()
        self._branch_lengths = new_branch_lengths
        self.update_hit_class_probs()

    def update_hit_class_probs(self):
        """Compute hit class probabilities for all codons in each sequence based on
        current branch lengths."""
        new_codon_probs = []
        new_hc_probs = []
        for (
            encoded_parent,
            nt_rates,
            nt_csps,
            branch_length,
        ) in zip(
            self.nt_parents,
            self.nt_ratess,
            self.nt_cspss,
            self.branch_lengths,
        ):
            scaled_rates = branch_length * nt_rates

            codon_probs = codon_probs_of_parent_scaled_nt_rates_and_csps(
                encoded_parent,
                scaled_rates[: len(encoded_parent)],
                nt_csps[: len(encoded_parent)],
            )
            new_codon_probs.append(codon_probs)

            new_hc_probs.append(
                hit_class_probs_tensor(reshape_for_codons(encoded_parent), codon_probs)
            )
        self.codon_probs = torch.stack(new_codon_probs)
        self.hit_class_probs = torch.stack(new_hc_probs)

    # A couple of these methods could maybe be moved to a super class, which itself subclasses Dataset
    def export_branch_lengths(self, out_csv_path):
        pd.DataFrame({"branch_length": self.branch_lengths}).to_csv(
            out_csv_path, index=False
        )

    def load_branch_lengths(self, in_csv_path):
        self.branch_lengths = torch.tensor(
            pd.read_csv(in_csv_path)["branch_length"].values
        )

    def __len__(self):
        return len(self.nt_parents)

    def __getitem__(self, idx):
        return {
            "parent": self.nt_parents[idx],
            "child": self.nt_children[idx],
            "observed_hcs": self.observed_hcs[idx],
            "nt_rates": self.nt_ratess[idx],
            "nt_csps": self.nt_cspss[idx],
            "hit_class_probs": self.hit_class_probs[idx],
            "codon_probs": self.codon_probs[idx],
            "codon_mask": self.codon_mask[idx],
        }

    def to(self, device):
        self.nt_parents = self.nt_parents.to(device)
        self.nt_children = self.nt_children.to(device)
        self.observed_hcs = self.observed_hcs.to(device)
        self.nt_ratess = self.nt_ratess.to(device)
        self.nt_cspss = self.nt_cspss.to(device)
        self.hit_class_probs = self.hit_class_probs.to(device)
        self.codon_mask = self.codon_mask.to(device)
        self.branch_lengths = self.branch_lengths.to(device)


def flatten_and_mask_sequence_codons(
    input_tensor: torch.Tensor, codon_mask: torch.Tensor = None
):
    """Flatten first dimension of input_tensor, applying codon_mask first if provided.

    This is useful for input_tensors whose first dimension represents sequences, and
    whose second dimension represents codons. The resulting tensor will then aggregate
    the codons of all sequences into the first dimension.
    """
    flat_input = input_tensor.flatten(start_dim=0, end_dim=1)
    if codon_mask is not None:
        flat_codon_mask = codon_mask.flatten()
        flat_input = flat_input[flat_codon_mask]
    return flat_input


def child_codon_probs_from_per_parent_probs(per_parent_probs, child_codon_idxs):
    """Calculate the probability of each child codon given the parent codon
    probabilities.

    Args:
        per_parent_probs (torch.Tensor): A (codon_count, 4, 4, 4) shaped tensor containing the probabilities
            of each possible target codon, for each parent codon.
        child_codon_idxs (torch.Tensor): A (codon_count, 3) shaped tensor containing the nucleotide indices for each child codon.

    Returns:
        torch.Tensor: A (codon_count,) shaped tensor containing the probabilities of each child codon.
    """
    return per_parent_probs[
        torch.arange(child_codon_idxs.size(0)),
        child_codon_idxs[:, 0],
        child_codon_idxs[:, 1],
        child_codon_idxs[:, 2],
    ]


def child_codon_probs_corrected(
    uncorrected_per_parent_probs, parent_codon_idxs, child_codon_idxs, model
):
    """Calculate the probability of each child codon given the parent codon
    probabilities, corrected by hit class factors.

    Args:
        uncorrected_per_parent_probs (torch.Tensor): A (codon_count, 4, 4, 4) shaped tensor containing the probabilities
            of each possible target codon, for each parent codon.
        parent_codon_idxs (torch.Tensor): A (codon_count, 3) shaped tensor containing the nucleotide indices for each parent codon
        child_codon_idxs (torch.Tensor): A (codon_count, 3) shaped tensor containing the nucleotide indices for each child codon
        model: A HitClassModel implementing the desired correction.

    Returns:
        torch.Tensor: A (codon_count,) shaped tensor containing the corrected probabilities of each child codon.
    """

    corrected_per_parent_probs = model(parent_codon_idxs, uncorrected_per_parent_probs)
    return child_codon_probs_from_per_parent_probs(
        corrected_per_parent_probs, child_codon_idxs
    )


class MultihitBurrito(Burrito):
    def __init__(
        self,
        train_dataset: HitClassDataset,
        val_dataset: HitClassDataset,
        model: HitClassModel,
        *args,
        **kwargs,
    ):
        super().__init__(
            train_dataset,
            val_dataset,
            model,
            *args,
            **kwargs,
        )

    def load_branch_lengths(self, in_csv_prefix):
        if self.train_loader is not None:
            self.train_loader.dataset.load_branch_lengths(
                in_csv_prefix + ".train_branch_lengths.csv"
            )
        self.val_loader.dataset.load_branch_lengths(
            in_csv_prefix + ".val_branch_lengths.csv"
        )

    def loss_of_batch(self, batch):
        child_idxs = batch["child"]
        parent_idxs = batch["parent"]
        codon_probs = batch["codon_probs"]
        codon_mask = batch["codon_mask"]

        # remove first dimension of parent_idxs by concatenating all of its elements
        codon_mask_flat = codon_mask.flatten(start_dim=0, end_dim=1)
        parent_codons_flat = reshape_for_codons(
            parent_idxs.flatten(start_dim=0, end_dim=1)
        )[codon_mask_flat]
        child_codons_flat = reshape_for_codons(
            child_idxs.flatten(start_dim=0, end_dim=1)
        )[codon_mask_flat]

        flat_masked_codon_probs = flatten_and_mask_sequence_codons(
            codon_probs, codon_mask=codon_mask
        )

        child_codon_logprobs = child_codon_probs_corrected(
            flat_masked_codon_probs,
            parent_codons_flat,
            child_codons_flat,
            self.model,
        ).log()
        return -child_codon_logprobs.sum()

    def _find_optimal_branch_length(
        self,
        parent_idxs,
        child_idxs,
        nt_rates,
        nt_csps,
        codon_mask,
        starting_branch_length,
        **optimization_kwargs,
    ):

        def log_pcp_probability(log_branch_length):
            # We want to first return the log-probability of the observed branch, using codon probs.
            # Then we'll want to adjust codon probs using our hit class probabilities
            branch_length = torch.exp(log_branch_length)
            scaled_rates = nt_rates * branch_length
            # Rates is a 1d tensor containing one rate for each nt site.

            # Codon probs is a Cx4x4x4 tensor containing for each codon idx the
            # distribution on possible target codons (all 64 of them!)
            codon_probs = codon_probs_of_parent_scaled_nt_rates_and_csps(
                parent_idxs, scaled_rates, nt_csps
            )[codon_mask]

            child_codon_idxs = reshape_for_codons(child_idxs)[codon_mask]
            parent_codon_idxs = reshape_for_codons(parent_idxs)[codon_mask]
            return (
                child_codon_probs_corrected(
                    codon_probs, parent_codon_idxs, child_codon_idxs, self.model
                )
                .log()
                .sum()
            )

        return optimize_branch_length(
            log_pcp_probability,
            starting_branch_length.double().item(),
            **optimization_kwargs,
        )[0]

    def find_optimal_branch_lengths(self, dataset, **optimization_kwargs):
        optimal_lengths = []

        for (
            parent_idxs,
            child_idxs,
            nt_rates,
            nt_csps,
            codon_mask,
            starting_length,
        ) in tqdm(
            zip(
                dataset.nt_parents,
                dataset.nt_children,
                dataset.nt_ratess,
                dataset.nt_cspss,
                dataset.codon_mask,
                dataset.branch_lengths,
            ),
            total=len(dataset.nt_parents),
            desc="Optimizing branch lengths",
        ):
            optimal_lengths.append(
                self._find_optimal_branch_length(
                    parent_idxs,
                    child_idxs,
                    nt_rates[: len(parent_idxs)],
                    nt_csps[: len(parent_idxs), :],
                    codon_mask,
                    starting_length,
                    **optimization_kwargs,
                )
            )

        return torch.tensor(optimal_lengths)

    def to_crepe(self):
        training_hyperparameters = {
            key: self.__dict__[key]
            for key in [
                "learning_rate",
            ]
        }
        encoder = framework.PlaceholderEncoder()
        return framework.Crepe(encoder, self.model, training_hyperparameters)


def hit_class_dataset_from_pcp_df(
    pcp_df: pd.DataFrame, branch_length_multiplier: int = 1.0
) -> HitClassDataset:
    nt_parents = pcp_df["parent"].reset_index(drop=True)
    nt_children = pcp_df["child"].reset_index(drop=True)
    nt_rates = pcp_df["nt_rates"].reset_index(drop=True)
    nt_csps = pcp_df["nt_csps"].reset_index(drop=True)

    return HitClassDataset(
        nt_parents,
        nt_children,
        nt_rates,
        nt_csps,
        branch_length_multiplier=branch_length_multiplier,
    )


def train_test_datasets_of_pcp_df(
    pcp_df: pd.DataFrame, train_frac: float = 0.8, branch_length_multiplier: float = 1.0
) -> Tuple[HitClassDataset, HitClassDataset]:
    """Splits a pcp_df prepared by `prepare_pcp_df` into a training and testing
    HitClassDataset."""
    nt_parents = pcp_df["parent"].reset_index(drop=True)
    nt_children = pcp_df["child"].reset_index(drop=True)
    nt_rates = pcp_df["nt_rates"].reset_index(drop=True)
    nt_csps = pcp_df["nt_csps"].reset_index(drop=True)

    train_len = int(train_frac * len(nt_parents))
    train_parents, val_parents = nt_parents[:train_len], nt_parents[train_len:]
    train_children, val_children = nt_children[:train_len], nt_children[train_len:]
    train_rates, val_rates = nt_rates[:train_len], nt_rates[train_len:]
    train_nt_csps, val_nt_csps = (
        nt_csps[:train_len],
        nt_csps[train_len:],
    )
    val_dataset = HitClassDataset(
        val_parents,
        val_children,
        val_rates,
        val_nt_csps,
        branch_length_multiplier=branch_length_multiplier,
    )
    if train_frac == 0.0:
        return None, val_dataset
    # else:
    train_dataset = HitClassDataset(
        train_parents,
        train_children,
        train_rates,
        train_nt_csps,
        branch_length_multiplier=branch_length_multiplier,
    )
    return val_dataset, train_dataset


def prepare_pcp_df(
    pcp_df: pd.DataFrame, crepe: framework.Crepe, site_count: int
) -> pd.DataFrame:
    """Trim parent and child sequences in pcp_df to codon boundaries and add the
    nt_rates and substitution probabilities.

    Returns the modified dataframe, which is the input dataframe modified in-place.
    """
    pcp_df["parent"] = _trim_to_codon_boundary_and_max_len(pcp_df["parent"], site_count)
    pcp_df["child"] = _trim_to_codon_boundary_and_max_len(pcp_df["child"], site_count)
    pcp_df = pcp_df[pcp_df["parent"] != pcp_df["child"]].reset_index(drop=True)
    ratess, cspss = framework.trimmed_shm_model_outputs_of_crepe(
        crepe, pcp_df["parent"]
    )
    pcp_df["nt_rates"] = ratess
    pcp_df["nt_csps"] = cspss
    return pcp_df



"""Defining the deep natural selection model (DNSM)."""

import copy

import torch
import torch.nn.functional as F

from netam.common import (
    clamp_probability,
)
from netam.dxsm import DXSMDataset, DXSMBurrito
from netam.hyper_burrito import HyperBurrito
import netam.molevol as molevol
import netam.sequences as sequences


class DNSMDataset(DXSMDataset):
    prefix = "dnsm"

    def update_neutral_probs(self):
        """Update the neutral mutation probabilities for the dataset.

        This is a somewhat vague name, but that's because it includes both the cases of
        the DNSM (in which case it's neutral probabilities of any nonsynonymous
        mutation) and the DASM (in which case it's the neutral probabilities of mutation
        to the various amino acids).

        This is the case of the DNSM, but the DASM will override this method.
        """
        neutral_aa_mut_prob_l = []

        for nt_parent, mask, nt_rates, nt_csps, branch_length in zip(
            self.nt_parents,
            self.masks,
            self.nt_ratess,
            self.nt_cspss,
            self._branch_lengths,
        ):
            mask = mask.to("cpu")
            nt_rates = nt_rates.to("cpu")
            nt_csps = nt_csps.to("cpu")
            if self.multihit_model is not None:
                multihit_model = copy.deepcopy(self.multihit_model).to("cpu")
            else:
                multihit_model = None
            # Note we are replacing all Ns with As, which means that we need to be careful
            # with masking out these positions later. We do this below.
            parent_idxs = sequences.nt_idx_tensor_of_str(nt_parent.replace("N", "A"))
            parent_len = len(nt_parent)
            # Cannot assume that nt_csps and mask are same length, because when
            # datasets are split, masks are recomputed.
            nt_mask = mask.repeat_interleave(3)[:parent_len]
            molevol.check_csps(parent_idxs[nt_mask], nt_csps[:parent_len][nt_mask])

            mut_probs = 1.0 - torch.exp(-branch_length * nt_rates[:parent_len])
            nt_csps = nt_csps[:parent_len, :]

            neutral_aa_mut_prob = molevol.neutral_aa_mut_probs(
                parent_idxs.reshape(-1, 3),
                mut_probs.reshape(-1, 3),
                nt_csps.reshape(-1, 3, 4),
                multihit_model=multihit_model,
            )

            if not torch.isfinite(neutral_aa_mut_prob).all():
                print(f"Found a non-finite neutral_aa_mut_prob")
                print(f"nt_parent: {nt_parent}")
                print(f"mask: {mask}")
                print(f"nt_rates: {nt_rates}")
                print(f"nt_csps: {nt_csps}")
                print(f"branch_length: {branch_length}")
                raise ValueError(
                    f"neutral_aa_mut_prob is not finite: {neutral_aa_mut_prob}"
                )

            # Ensure that all values are positive before taking the log later
            neutral_aa_mut_prob = clamp_probability(neutral_aa_mut_prob)

            pad_len = self.max_aa_seq_len - neutral_aa_mut_prob.shape[0]
            if pad_len > 0:
                neutral_aa_mut_prob = F.pad(
                    neutral_aa_mut_prob, (0, pad_len), value=1e-8
                )
            # Here we zero out masked positions.
            neutral_aa_mut_prob *= mask

            neutral_aa_mut_prob_l.append(neutral_aa_mut_prob)

        # Note that our masked out positions will have a nan log probability,
        # which will require us to handle them correctly downstream.
        self.log_neutral_aa_mut_probss = torch.log(torch.stack(neutral_aa_mut_prob_l))

    def __getitem__(self, idx):
        return {
            "aa_parents_idxs": self.aa_parents_idxss[idx],
            "aa_subs_indicator": self.aa_subs_indicators[idx],
            "mask": self.masks[idx],
            "log_neutral_aa_mut_probs": self.log_neutral_aa_mut_probss[idx],
            "nt_rates": self.nt_ratess[idx],
            "nt_csps": self.nt_cspss[idx],
        }

    def to(self, device):
        self.aa_parents_idxss = self.aa_parents_idxss.to(device)
        self.aa_subs_indicators = self.aa_subs_indicators.to(device)
        self.masks = self.masks.to(device)
        self.log_neutral_aa_mut_probss = self.log_neutral_aa_mut_probss.to(device)
        self.nt_ratess = self.nt_ratess.to(device)
        self.nt_cspss = self.nt_cspss.to(device)
        if self.multihit_model is not None:
            self.multihit_model = self.multihit_model.to(device)


class DNSMBurrito(DXSMBurrito):
    prefix = "dnsm"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def prediction_pair_of_batch(self, batch):
        """Get log neutral amino acid substitution probabilities and log selection
        factors for a batch of data."""
        aa_parents_idxs = batch["aa_parents_idxs"].to(self.device)
        mask = batch["mask"].to(self.device)
        log_neutral_aa_mut_probs = batch["log_neutral_aa_mut_probs"].to(self.device)
        if not torch.isfinite(log_neutral_aa_mut_probs[mask]).all():
            raise ValueError(
                f"log_neutral_aa_mut_probs has non-finite values at relevant positions: {log_neutral_aa_mut_probs[mask]}"
            )
        log_selection_factors = self.model(aa_parents_idxs, mask)
        return log_neutral_aa_mut_probs, log_selection_factors

    def predictions_of_pair(self, log_neutral_aa_mut_probs, log_selection_factors):
        """Obtain the predictions for a pair consisting of the log neutral amino acid
        mutation substitution probabilities and the log selection factors."""
        predictions = torch.exp(log_neutral_aa_mut_probs + log_selection_factors)
        assert torch.isfinite(predictions).all()
        predictions = clamp_probability(predictions)
        return predictions

    def predictions_of_batch(self, batch):
        """Make predictions for a batch of data.

        Note that we use the mask for prediction as part of the input for the
        transformer, though we don't mask the predictions themselves.
        """
        log_neutral_aa_mut_probs, log_selection_factors = self.prediction_pair_of_batch(
            batch
        )
        return self.predictions_of_pair(log_neutral_aa_mut_probs, log_selection_factors)

    def loss_of_batch(self, batch):
        aa_subs_indicator = batch["aa_subs_indicator"].to(self.device)
        mask = batch["mask"].to(self.device)
        aa_subs_indicator = aa_subs_indicator.masked_select(mask)
        predictions = self.predictions_of_batch(batch).masked_select(mask)
        return self.bce_loss(predictions, aa_subs_indicator)

    def build_selection_matrix_from_parent(self, parent: str):
        """Build a selection matrix from a parent amino acid sequence.

        Values at ambiguous sites are meaningless.
        """
        parent = sequences.translate_sequence(parent)
        selection_factors = self.model.selection_factors_of_aa_str(parent)
        selection_matrix = torch.zeros((len(selection_factors), 20), dtype=torch.float)
        # Every "off-diagonal" entry of the selection matrix is set to the selection
        # factor, where "diagonal" means keeping the same amino acid.
        selection_matrix[:, :] = selection_factors[:, None]
        parent = parent.replace("X", "A")
        # Set "diagonal" elements to one.
        parent_idxs = sequences.aa_idx_array_of_str(parent)
        selection_matrix[torch.arange(len(parent_idxs)), parent_idxs] = 1.0

        return selection_matrix


class DNSMHyperBurrito(HyperBurrito):
    # Note that we have to write the args out explicitly because we use some magic to filter kwargs in the optuna_objective method.
    def burrito_of_model(
        self,
        model,
        device,
        batch_size=1024,
        learning_rate=0.1,
        min_learning_rate=1e-4,
        weight_decay=1e-6,
    ):
        model.to(device)
        burrito = DNSMBurrito(
            self.train_dataset,
            self.val_dataset,
            model,
            batch_size=batch_size,
            learning_rate=learning_rate,
            min_learning_rate=min_learning_rate,
            weight_decay=weight_decay,
        )
        return burrito



"""Code for handling sequences and sequence files."""

import itertools

import torch
import numpy as np

from Bio import SeqIO
from Bio.Seq import Seq

AA_STR_SORTED = "ACDEFGHIKLMNPQRSTVWY"
NT_STR_SORTED = "ACGT"
CODONS = [
    "".join(codon_list)
    for codon_list in itertools.product(["A", "C", "G", "T"], repeat=3)
]
STOP_CODONS = ["TAA", "TAG", "TGA"]


def nt_idx_array_of_str(nt_str):
    """Return the indices of the nucleotides in a string."""
    try:
        return np.array([NT_STR_SORTED.index(nt) for nt in nt_str])
    except ValueError:
        print(f"Found an invalid nucleotide in the string: {nt_str}")
        raise


def aa_idx_array_of_str(aa_str):
    """Return the indices of the amino acids in a string."""
    try:
        return np.array([AA_STR_SORTED.index(aa) for aa in aa_str])
    except ValueError:
        print(f"Found an invalid amino acid in the string: {aa_str}")
        raise


def nt_idx_tensor_of_str(nt_str):
    """Return the indices of the nucleotides in a string."""
    try:
        return torch.tensor([NT_STR_SORTED.index(nt) for nt in nt_str])
    except ValueError:
        print(f"Found an invalid nucleotide in the string: {nt_str}")
        raise


def aa_idx_tensor_of_str(aa_str):
    """Return the indices of the amino acids in a string."""
    try:
        return torch.tensor([AA_STR_SORTED.index(aa) for aa in aa_str])
    except ValueError:
        print(f"Found an invalid amino acid in the string: {aa_str}")
        raise


def aa_onehot_tensor_of_str(aa_str):
    aa_onehot = torch.zeros((len(aa_str), 20))
    aa_indices_parent = aa_idx_array_of_str(aa_str)
    aa_onehot[torch.arange(len(aa_str)), aa_indices_parent] = 1
    return aa_onehot


def generic_subs_indicator_tensor_of(ambig_symb, parent, child):
    """Return a tensor indicating which positions in the parent sequence are substituted
    in the child sequence."""
    return torch.tensor(
        [
            0 if (p == ambig_symb or c == ambig_symb) else p != c
            for p, c in zip(parent, child)
        ],
        dtype=torch.float,
    )


def nt_subs_indicator_tensor_of(parent, child):
    """Return a tensor indicating which positions in the parent sequence are substituted
    in the child sequence."""
    return generic_subs_indicator_tensor_of("N", parent, child)


def aa_subs_indicator_tensor_of(parent, child):
    """Return a tensor indicating which positions in the parent sequence are substituted
    in the child sequence."""
    return generic_subs_indicator_tensor_of("X", parent, child)


def read_fasta_sequences(file_path):
    with open(file_path, "r") as handle:
        sequences = [str(record.seq) for record in SeqIO.parse(handle, "fasta")]
    return sequences


def translate_sequences(nt_sequences):
    aa_sequences = []
    for seq in nt_sequences:
        if len(seq) % 3 != 0:
            raise ValueError(f"The sequence '{seq}' is not a multiple of 3.")
        aa_seq = str(Seq(seq).translate())
        if "*" in aa_seq:
            raise ValueError(f"The sequence '{seq}' contains a stop codon.")
        aa_sequences.append(aa_seq)
    return aa_sequences


def translate_sequence(nt_sequence):
    return translate_sequences([nt_sequence])[0]


def aa_index_of_codon(codon):
    """Return the index of the amino acid encoded by a codon."""
    aa = translate_sequence(codon)
    return AA_STR_SORTED.index(aa)


def generic_mutation_frequency(ambig_symb, parent, child):
    """Return the fraction of sites that differ between the parent and child
    sequences."""
    return sum(
        1
        for p, c in zip(parent, child)
        if p != c and p != ambig_symb and c != ambig_symb
    ) / len(parent)


def nt_mutation_frequency(parent, child):
    """Return the fraction of nucleotide sites that differ between the parent and child
    sequences."""
    return generic_mutation_frequency("N", parent, child)


def aa_mutation_frequency(parent, child):
    """Return the fraction of amino acid sites that differ between the parent and child
    sequences."""
    return generic_mutation_frequency("X", parent, child)


def assert_pcp_lengths(parent, child):
    """Assert that the lengths of the parent and child sequences are the same and that
    they are multiples of 3."""
    if len(parent) != len(child):
        raise ValueError(
            f"The parent and child sequences are not the same length: "
            f"{len(parent)} != {len(child)}"
        )
    if len(parent) % 3 != 0:
        raise ValueError(f"Found a PCP with length not a multiple of 3: {len(parent)}")


def pcp_criteria_check(parent, child, max_mut_freq=0.3):
    """Check that parent child pair undergoes mutation at a reasonable rate."""
    if parent == child:
        return False
    elif nt_mutation_frequency(parent, child) > max_mut_freq:
        return False
    else:
        return True


def generate_codon_aa_indicator_matrix():
    """Generate a matrix that maps codons (rows) to amino acids (columns)."""

    matrix = np.zeros((len(CODONS), len(AA_STR_SORTED)))

    for i, codon in enumerate(CODONS):
        try:
            aa = translate_sequences([codon])[0]
            aa_idx = AA_STR_SORTED.index(aa)
            matrix[i, aa_idx] = 1
        except ValueError:  # Handle STOP codon
            pass

    return matrix


CODON_AA_INDICATOR_MATRIX = torch.tensor(
    generate_codon_aa_indicator_matrix(), dtype=torch.float32
)


def assert_full_sequences(parent, child):
    """Assert that the parent and child sequences full length, containing no ambiguous
    bases (N)."""

    if "N" in parent or "N" in child:
        raise ValueError("Found ambiguous bases in the parent or child sequence.")


def apply_aa_mask_to_nt_sequence(nt_seq, aa_mask):
    """Apply an amino acid mask to a nucleotide sequence."""
    return "".join(
        nt for nt, mask_val in zip(nt_seq, aa_mask.repeat_interleave(3)) if mask_val
    )


def iter_codons(nt_seq):
    """Iterate over the codons in a nucleotide sequence."""
    for i in range(0, (len(nt_seq) // 3) * 3, 3):
        yield nt_seq[i : i + 3]



"""HyperBurrito class for optimizing hyperparameters."""

from abc import ABC, abstractmethod
from datetime import datetime
import inspect

import optuna

from netam.common import parameter_count_of_model
from netam.framework import SHMBurrito


def filter_kwargs(func, kwargs):
    """Filter kwargs to only those that the function accepts."""
    # Get the parameters of the function
    sig = inspect.signature(func)
    func_params = sig.parameters

    # Filter kwargs to only those that the function accepts
    return {k: v for k, v in kwargs.items() if k in func_params}


def timestamp_str():
    return datetime.now().strftime("%Y%m%d_%H%M%S")


class HyperBurrito(ABC):
    """A class to optimize hyperparameters."""

    def __init__(
        self,
        device,
        train_dataset,
        val_dataset,
        model_class,
        epochs=100,
    ):
        self.device = device
        self.train_dataset = train_dataset
        self.val_dataset = val_dataset
        train_dataset.to(self.device)
        val_dataset.to(self.device)
        self.model_class = model_class
        self.epochs = epochs

    @abstractmethod
    def burrito_of_model(self, model, **kwargs):
        pass

    def optuna_objective(
        self,
        trial,
        int_params,
        cat_params,
        float_params,
        log_float_params,
        fixed_hyperparams=None,
    ):
        """Optuna objective function.

        Return validation loss unless the model has more parameters than allowed
        in fixed_hyperparams["max_parameter_count"], in which case return 1e9.

        Note that if a parameter appears in one of the xxx_params dictionaries
        used for sampling as well as the fixed_hyperparams dictionary, the
        sampled value will be used.

        Args:
            trial (optuna.Trial): Optuna trial object.
            int_params (dict): Dictionary of integer parameters to optimize.
            cat_params (dict): Dictionary of categorical parameters to optimize.
            float_params (dict): Dictionary of float parameters to optimize on a linear scale.
            log_float_params (dict): Dictionary of float parameters to optimize on a log scale.
            fixed_hyperparams (dict, optional): Dictionary of fixed hyperparameters. Defaults to None.

        Returns:
            float: Validation loss or 1e9 if the model has too many parameters.
        """
        if fixed_hyperparams is not None:
            for key, value in fixed_hyperparams.items():
                trial.set_user_attr(key, value)

        hyperparams = fixed_hyperparams or {}

        for param_name, choices in cat_params.items():
            hyperparams[param_name] = trial.suggest_categorical(param_name, choices)

        for param_name, (low, high) in int_params.items():
            hyperparams[param_name] = trial.suggest_int(param_name, low, high)

        for param_name, (low, high) in float_params.items():
            hyperparams[param_name] = trial.suggest_float(param_name, low, high)

        for param_name, (low, high) in log_float_params.items():
            hyperparams[param_name] = trial.suggest_float(
                param_name, low, high, log=True
            )

        model_hyperparams = filter_kwargs(self.model_class, hyperparams)
        model = self.model_class(**model_hyperparams)
        model.to(self.device)

        if hyperparams is not None and "max_parameter_count" in hyperparams:
            parameter_count = parameter_count_of_model(model)
            # if parameter_count is not in the range between hyperparams["min_parameter_count"] and hyperparams["max_parameter_count"]:
            if parameter_count not in range(
                hyperparams["min_parameter_count"],
                hyperparams["max_parameter_count"] + 1,
            ):
                range_str = f"[{hyperparams['min_parameter_count']}, {hyperparams['max_parameter_count']}]"
                print(
                    f"Trial rejected. Model has {parameter_count} parameters, not in {range_str}]."
                )
                return 1e9

        burrito_hyperparams = filter_kwargs(self.burrito_of_model, hyperparams)
        print("burrito_hypers:", burrito_hyperparams)
        burrito = self.burrito_of_model(model, **burrito_hyperparams)

        losses = burrito.joint_train(epochs=self.epochs)

        return losses["val_loss"].min()

    def optuna_optimize(
        self,
        n_trials,
        cat_params,
        int_params,
        float_params,
        log_float_params,
        fixed_hyperparams=None,
        study_name=None,
    ):
        storage_url = "sqlite:///_ignore/optuna.db"
        if study_name is None:
            study_name = f"study_{self.model_class.__name__}"

        # Create or load the study
        study = optuna.create_study(
            direction="minimize",
            study_name=study_name,
            storage=storage_url,
            load_if_exists=True,
        )

        study.optimize(
            lambda trial: self.optuna_objective(
                trial,
                int_params,
                cat_params,
                float_params,
                log_float_params,
                fixed_hyperparams,
            ),
            n_trials=n_trials,
        )
        best_hyperparams = study.best_params
        best_score = study.best_value
        print(f"Best Hyperparameters: {best_hyperparams}")
        print(f"Best Validation Loss: {best_score}")

        output_path = (
            f"_ignore/optuna_{self.model_class.__name__}_{timestamp_str()}.csv"
        )
        trial_data = study.trials_dataframe()
        trial_data.to_csv(output_path, index=False)


class SHMHyperBurrito(HyperBurrito):
    # Note that we have to write the args out explicitly because we use some
    # magic to filter kwargs in the optuna_objective method.
    def burrito_of_model(
        self,
        model,
        batch_size=1024,
        learning_rate=0.1,
        min_learning_rate=1e-4,
        weight_decay=1e-6,
    ):
        burrito = SHMBurrito(
            self.train_dataset,
            self.val_dataset,
            model,
            batch_size=batch_size,
            learning_rate=learning_rate,
            min_learning_rate=min_learning_rate,
            weight_decay=weight_decay,
        )
        return burrito



import os

import torch
import pytest

from netam.common import BIG, force_spawn
from netam.framework import (
    crepe_exists,
    load_crepe,
)
from netam.models import TransformerBinarySelectionModelWiggleAct
from netam.dasm import (
    DASMBurrito,
    DASMDataset,
    zap_predictions_along_diagonal,
)


@pytest.fixture(scope="module")
def dasm_burrito(pcp_df):
    force_spawn()
    """Fixture that returns the DNSM Burrito object."""
    pcp_df["in_train"] = True
    pcp_df.loc[pcp_df.index[-15:], "in_train"] = False
    train_dataset, val_dataset = DASMDataset.train_val_datasets_of_pcp_df(pcp_df)

    model = TransformerBinarySelectionModelWiggleAct(
        nhead=2,
        d_model_per_head=4,
        dim_feedforward=256,
        layer_count=2,
        output_dim=20,
    )

    burrito = DASMBurrito(
        train_dataset,
        val_dataset,
        model,
        batch_size=32,
        learning_rate=0.001,
        min_learning_rate=0.0001,
    )
    burrito.joint_train(
        epochs=1, cycle_count=2, training_method="full", optimize_bl_first_cycle=False
    )
    return burrito


def test_parallel_branch_length_optimization(dasm_burrito):
    dataset = dasm_burrito.val_dataset
    parallel_branch_lengths = dasm_burrito.find_optimal_branch_lengths(dataset)
    branch_lengths = dasm_burrito.serial_find_optimal_branch_lengths(dataset)
    assert torch.allclose(branch_lengths, parallel_branch_lengths)


def test_crepe_roundtrip(dasm_burrito):
    os.makedirs("_ignore", exist_ok=True)
    crepe_path = "_ignore/dasm"
    dasm_burrito.save_crepe(crepe_path)
    assert crepe_exists(crepe_path)
    crepe = load_crepe(crepe_path)
    model = crepe.model
    assert isinstance(model, TransformerBinarySelectionModelWiggleAct)
    assert dasm_burrito.model.hyperparameters == model.hyperparameters
    model.to(dasm_burrito.device)
    for t1, t2 in zip(
        dasm_burrito.model.state_dict().values(), model.state_dict().values()
    ):
        assert torch.equal(t1, t2)


def test_zap_diagonal(dasm_burrito):
    batch = dasm_burrito.val_dataset[0:2]
    predictions = dasm_burrito.predictions_of_batch(batch)
    predictions = torch.cat(
        [predictions, torch.zeros_like(predictions[:, :, :1])], dim=-1
    )
    aa_parents_idxs = batch["aa_parents_idxs"].to(dasm_burrito.device)
    zeroed_predictions = predictions.clone()
    zeroed_predictions = zap_predictions_along_diagonal(
        zeroed_predictions, aa_parents_idxs
    )
    L = predictions.shape[1]
    for batch_idx in range(2):
        for i in range(L):
            for j in range(20):
                if j == aa_parents_idxs[batch_idx, i]:
                    assert zeroed_predictions[batch_idx, i, j] == -BIG
                else:
                    assert (
                        zeroed_predictions[batch_idx, i, j]
                        == predictions[batch_idx, i, j]
                    )



import torch
import numpy as np


# Define the number of bases (e.g., 4 for DNA/RNA)
_num_bases = 4

# Generate all possible codons using broadcasting
_i, _j, _k = np.indices((_num_bases, _num_bases, _num_bases))  # Create index grids
_codon1 = np.stack((_i, _j, _k), axis=-1)  # Shape: (4, 4, 4, 3)

# Expand dimensions to compare all codon pairs using broadcasting
_codon1_expanded = _codon1[
    :, :, :, np.newaxis, np.newaxis, np.newaxis, :
]  # Shape: (4, 4, 4, 1, 1, 1, 3)
_codon2_expanded = _codon1[
    np.newaxis, np.newaxis, np.newaxis, :, :, :, :
]  # Shape: (1, 1, 1, 4, 4, 4, 3)

# Count the number of differing positions between each pair of codons
"""hit_class_tensor is a tensor of shape (4, 4, 4, 4, 4, 4) recording the hit class (number of nucleotide differences) between all possible parent and child codons. The first three dimensions represent the parent codon, and the last three represent the child codon. Codons are identified by triples of nucleotide indices from `netam.common.BASES`."""
hit_class_tensor = torch.tensor(
    np.sum(_codon1_expanded != _codon2_expanded, axis=-1)
).int()


def parent_specific_hit_classes(parent_codon_idxs: torch.Tensor) -> torch.Tensor:
    """Produce a tensor containing the hit classes of all possible child codons, for
    each passed parent codon.

    Args:
        parent_codon_idxs (torch.Tensor): A (codon_count, 3) shaped tensor containing for each codon, the
            indices of the parent codon's nucleotides.
    Returns:
        torch.Tensor: A (codon_count, 4, 4, 4) shaped tensor containing the hit classes of each possible child codon for each parent codon.
    """
    return hit_class_tensor[
        parent_codon_idxs[:, 0], parent_codon_idxs[:, 1], parent_codon_idxs[:, 2]
    ]


def apply_multihit_correction(
    parent_codon_idxs: torch.Tensor,
    codon_probs: torch.Tensor,
    log_hit_class_factors: torch.Tensor,
) -> torch.Tensor:
    """Multiply codon probabilities by their hit class factors, and renormalize.

    Suppose there are N codons, then the parameters are as follows:

    Args:
        parent_codon_idxs (torch.Tensor): A (N, 3) shaped tensor containing for each codon, the
            indices of the parent codon's nucleotides.
        codon_probs (torch.Tensor): A (N, 4, 4, 4) shaped tensor containing the probabilities
            of mutating to each possible target codon, for each of the N parent codons.
        log_hit_class_factors (torch.Tensor): A tensor containing the log hit class factors for hit classes 1, 2, and 3. The
            factor for hit class 0 is assumed to be 1 (that is, 0 in log-space).

    Returns:
        torch.Tensor: A (N, 4, 4, 4) shaped tensor containing the probabilities of mutating to each possible
            target codon, for each of the N parent codons, after applying the hit class factors.
    """
    per_parent_hit_class = parent_specific_hit_classes(parent_codon_idxs)
    corrections = torch.cat([torch.tensor([0.0]), log_hit_class_factors]).exp()
    reshaped_corrections = corrections[per_parent_hit_class]
    unnormalized_corrected_probs = codon_probs * reshaped_corrections
    normalizations = torch.sum(
        unnormalized_corrected_probs, dim=[1, 2, 3], keepdim=True
    )
    result = unnormalized_corrected_probs / normalizations
    if torch.any(torch.isnan(result)):
        print("NAN found in multihit correction application")
        assert False
    return result


def hit_class_probs_tensor(
    parent_codon_idxs: torch.Tensor, codon_probs: torch.Tensor
) -> torch.Tensor:
    """Calculate probabilities of hit classes between parent codons and all other codons
    for all the sites of a sequence.

    Args:
        parent_codon_idxs (torch.Tensor): The parent nucleotide sequence encoded as a tensor of shape (codon_count, 3),
            containing the nt indices of each codon.
        codon_probs (torch.Tensor): A (codon_count, 4, 4, 4) shaped tensor containing the probabilities of various
            codons, for each codon in parent seq.

    Returns:
        probs (torch.Tensor): A tensor containing the probabilities of different
            counts of hit classes between parent codons and
            all other codons, with shape (codon_count, 4).

    Notes:
    Uses hit_class_tensor (torch.Tensor): A 4x4x4x4x4x4 tensor which when indexed with a parent codon produces
    the hit classes to all possible child codons.
    """

    # Get a codon_countx4x4x4 tensor describing for each parent codon the hit classes of all child codons
    per_parent_hit_class = parent_specific_hit_classes(parent_codon_idxs)
    codon_count = per_parent_hit_class.size(0)
    hc_prob_tensor = torch.zeros(codon_count, 4)
    for k in range(4):
        mask = per_parent_hit_class == k
        hc_prob_tensor[:, k] = (codon_probs * mask).sum(dim=(1, 2, 3))

    return hc_prob_tensor
